import requests
import json
import gzip
import re
import traceback
import pprint

from requests.auth import HTTPBasicAuth
from unidecode import unidecode
from typing import List, Tuple, Any

valid_context_types = {'i', 'q', 'e', 'm', 'd'}


class ElasticsearchManager(object):
    @staticmethod
    def build_kgtk_search_input(kgtk_file_path: str,
                                label_fields: str,
                                mapping_file_path: str,
                                output_path: str,
                                alias_fields: str = None,
                                extra_alias_properties: str = None,
                                pagerank_fields: str = None,
                                add_text: bool = False,
                                description_properties: str = None,
                                separate_languages: bool = True,
                                property_datatype_file: str = None,
                                languages: set = None
                                ):
        """
        builds a json lines file and a mapping file to support retrieval of candidates
        It is assumed that the file is sorted by subject and predicate, in order to be able to process it in a
        streaming fashion

        Args:
        :param kgtk_file_path: input KGTK edge file
        :param label_fields: comma separated properties to be used as labels
        :param mapping_file_path: output file path for mapping json file
        :param output_path: output json lines file path
        :param alias_fields: comma separated properties to be used as aliases
        :param extra_alias_properties: additional properties to be used as aliases
        :param pagerank_fields: comma separated properties to be used as pagerank
        :param add_text: concatenate english labels, aliases and descriptions in one text field
        :param description_properties: comma separated properties to be used as descriptions
        :param separate_languages: flag to store text in separate languages in different fields
        :param property_datatype_file: input file with property datatype information
        :param languages: a set of languages, for labels, aliases and descriptions
        :return: None
        """

        if languages is None:
            languages = {'en'}
        labels = label_fields.split(',')
        aliases = alias_fields.split(',') if alias_fields else []
        extra_aliases_properties = set(extra_alias_properties.split(',')) if extra_alias_properties else set()
        pagerank = pagerank_fields.split(',') if pagerank_fields else []
        descriptions = description_properties.split(',') if description_properties else []

        # create the property data type dict
        property_datatype_dict = {}
        if property_datatype_file:
            property_datatype_dict = ElasticsearchManager.create_property_metadata_dict(property_datatype_file)

        if kgtk_file_path.endswith(".gz"):
            kgtk_file = gzip.open(kgtk_file_path)
        else:
            kgtk_file = open(kgtk_file_path, "r")

        if output_path.endswith(".gz"):
            output_file = gzip.open(output_path, mode='wt')
        else:
            output_file = open(output_path, 'w')

        _labels = dict()
        _aliases = dict()
        _descriptions = dict()
        _instance_ofs = set()
        data_type = None
        all_langs = set()
        lang = 'en'
        qnode_statement_count = 0
        is_class = False
        _wikitable_anchor_text = {}
        _wikipedia_anchor_text = {}
        _abbreviated_name = {}
        _redirect_text = {}
        _text_embedding = None
        _graph_embeddings_complex = None
        _graph_embeddings_transE = None
        ascii_labels = set()
        all_labels = set()
        property_count = None
        class_count = None
        context = None
        is_human = False
        _extra_aliases = set()
        _properties = set()
        _external_identifiers = set()
        _external_identifiers_pairs = set()

        _pagerank = 0.0

        prev_node = None
        i = 0

        column_header_dict = None
        try:
            for line in kgtk_file:
                i += 1
                if i % 1000000 == 0:
                    print('Processed {} lines...'.format(i))
                if isinstance(line, bytes):
                    line = line.decode('utf-8')
                line = line.replace('\n', '')
                if column_header_dict is None and 'node1' in line and 'id' in line and 'node2' in line:
                    # header line
                    cols = line.replace('\n', '').split('\t')
                    column_header_dict = {
                        'node1': cols.index('node1'),
                        'label': cols.index('label'),
                        'node2': cols.index('node2')
                    }

                # if line.startswith('Q'):
                else:
                    vals = line.split('\t')
                    node1_id = column_header_dict['node1']
                    label_id = column_header_dict['label']
                    node2_id = column_header_dict['node2']
                    node1 = vals[node1_id]
                    if '-' not in node1:  # ignore qualifiers
                        if prev_node is None:
                            prev_node = node1
                        if node1 != prev_node:
                            ElasticsearchManager._write_one_node(_labels=_labels,
                                                                 _aliases=_aliases,
                                                                 _pagerank=_pagerank,
                                                                 prev_node=prev_node,
                                                                 output_file=output_file,
                                                                 _descriptions=_descriptions,
                                                                 add_all_text=add_text,
                                                                 data_type=data_type,
                                                                 instance_ofs=_instance_ofs,
                                                                 qnode_statement_count=qnode_statement_count,
                                                                 is_class=is_class,
                                                                 wikitable_anchor_text=_wikitable_anchor_text,
                                                                 wikipedia_anchor_text=_wikipedia_anchor_text,
                                                                 abbreviated_name=_abbreviated_name,
                                                                 redirect_text=_redirect_text,
                                                                 text_embedding=_text_embedding,
                                                                 graph_embeddings_complex=_graph_embeddings_complex,
                                                                 graph_embeddings_transe=_graph_embeddings_transE,
                                                                 ascii_labels=ascii_labels,
                                                                 property_count=property_count,
                                                                 class_count=class_count,
                                                                 context=context,
                                                                 is_human=is_human,
                                                                 extra_aliases=_extra_aliases,
                                                                 properties=_properties,
                                                                 external_identifiers=_external_identifiers,
                                                                 external_identifiers_pairs=_external_identifiers_pairs
                                                                 )
                            # initialize for next node
                            _labels = dict()
                            _aliases = dict()
                            _descriptions = dict()
                            _instance_ofs = set()
                            data_type = None
                            _pagerank = 0.0
                            prev_node = node1
                            lang = 'en'
                            qnode_statement_count = 0
                            is_class = False
                            _wikitable_anchor_text = {}
                            _wikipedia_anchor_text = {}
                            _abbreviated_name = {}
                            _redirect_text = {}
                            _text_embedding = None
                            _graph_embeddings_complex = None
                            _graph_embeddings_transE = None
                            ascii_labels = set()
                            all_labels = set()
                            property_count = None
                            class_count = None
                            context = None
                            is_human = False
                            _extra_aliases = set()
                            _properties = set()
                            _external_identifiers = set()
                            _external_identifiers_pairs = set()

                        qnode_statement_count += 1
                        if vals[label_id] in labels:
                            if separate_languages:
                                tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            else:
                                tmp_val = ElasticsearchManager.remove_language_tag(vals[node2_id])
                            if lang in languages:
                                if lang not in _labels:
                                    _labels[lang] = set()
                                    all_langs.add(lang)

                                if tmp_val.strip() != '':
                                    _labels[lang].add(tmp_val)
                                    all_labels.add(tmp_val)

                                    if lang in {'en', 'de', 'es', 'fr', 'it', 'pt'}:
                                        # add transilerated value as well
                                        _ascii_label = ElasticsearchManager.transliterate_label(tmp_val)
                                        if _ascii_label != "" and _ascii_label not in all_labels:
                                            ascii_labels.add(_ascii_label)

                        elif vals[label_id] in aliases:
                            if separate_languages:
                                tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            else:
                                tmp_val = ElasticsearchManager.remove_language_tag(vals[node2_id])
                            if lang in languages:
                                if lang not in _aliases:
                                    _aliases[lang] = set()
                                    all_langs.add(lang)

                                if tmp_val.strip() != '':
                                    _aliases[lang].add(tmp_val)
                                    all_labels.add(tmp_val)

                                    if lang in {'en', 'de', 'es', 'fr', 'it', 'pt'}:
                                        # add transilerated value as well
                                        _ascii_alias = ElasticsearchManager.transliterate_label(tmp_val)
                                        if _ascii_alias != "" and _ascii_alias not in all_labels:
                                            ascii_labels.add(_ascii_alias)

                        elif vals[label_id] in pagerank:
                            tmp_val = ElasticsearchManager.to_float(vals[node2_id])
                            if tmp_val:
                                _pagerank = tmp_val
                        elif vals[label_id] in descriptions:
                            if separate_languages:
                                tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            else:
                                tmp_val = ElasticsearchManager.remove_language_tag(vals[node2_id])
                            if lang in languages:
                                if lang not in _descriptions:
                                    _descriptions[lang] = set()
                                    all_langs.add(lang)
                                if tmp_val.strip() != '':
                                    _descriptions[lang].add(tmp_val)
                        elif vals[label_id].strip() == 'isa_star' or vals[label_id].strip() == 'P39' \
                                or vals[label_id].strip() == 'P106' or vals[label_id].strip() == 'Pisa_star':
                            _instance_ofs.add(vals[node2_id])
                        elif vals[label_id].strip() == 'datatype':
                            data_type = vals[node2_id]
                        elif node1 in property_datatype_dict:
                            data_type = property_datatype_dict[node1]
                        elif vals[label_id] == 'P279' and vals[node2_id].startswith('Q'):
                            is_class = True
                        elif vals[label_id] == 'wikipedia_table_anchor':
                            tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            if tmp_val.strip() != "":
                                if lang not in _wikitable_anchor_text:
                                    _wikitable_anchor_text[lang] = set()
                                _wikitable_anchor_text[lang].add(tmp_val)
                        elif vals[label_id] == 'wikipedia_anchor':
                            tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            if tmp_val.strip() != "":
                                if lang not in _wikipedia_anchor_text:
                                    _wikipedia_anchor_text[lang] = set()
                                _wikipedia_anchor_text[lang].add(tmp_val)
                        elif vals[label_id] == 'redirect_from':
                            tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            if tmp_val.strip() != "":
                                if lang not in _redirect_text:
                                    _redirect_text[lang] = set()
                                _redirect_text[lang].add(tmp_val)
                        elif vals[label_id] == 'abbreviated_name':
                            tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            if tmp_val.strip() != "":
                                if lang not in _abbreviated_name:
                                    _abbreviated_name[lang] = set()
                                _abbreviated_name[lang].add(tmp_val)
                        elif vals[label_id] == 'graph_embeddings_complEx':
                            _graph_embeddings_complex = vals[node2_id]
                        elif vals[label_id] == 'graph_embeddings_transE':
                            _graph_embeddings_transE = vals[node2_id]
                        elif vals[label_id] == 'text_embedding':
                            _text_embedding = vals[node2_id]
                        elif vals[label_id] == 'property_count':
                            property_count = vals[node2_id]
                        elif vals[label_id] == 'class_count':
                            class_count = vals[node2_id]
                        elif vals[label_id] == 'context':
                            context_dict = ElasticsearchManager.parse_context_string(node1, vals[node2_id])
                            context = context_dict[node1]
                        elif vals[label_id] == 'P31' and vals[node2_id] == 'Q5':
                            is_human = True
                        elif vals[label_id] in extra_aliases_properties:
                            if separate_languages:
                                tmp_val, lang = ElasticsearchManager.separate_language_text_tag(vals[node2_id])
                            else:
                                tmp_val = ElasticsearchManager.remove_language_tag(vals[node2_id])

                            if lang in languages:
                                if tmp_val.strip() != '':
                                    _extra_aliases.add(tmp_val)

                                    if lang in {'en', 'de', 'es', 'fr', 'it', 'pt'}:
                                        _ascii_label = ElasticsearchManager.transliterate_label(tmp_val)
                                        if _ascii_label != "":
                                            _extra_aliases.add(_ascii_label)

                        if vals[label_id].startswith('P'):  # add to set of properties
                            _properties.add(vals[label_id])

                        if property_datatype_dict.get(vals[label_id], None) == 'external-id':
                            ex_id = vals[node2_id].replace('"', "").strip()
                            _external_identifiers.add(ex_id)
                            _external_identifiers_pairs.add(f"{vals[label_id]}:{ex_id}")

            # do one more write for last node
            ElasticsearchManager._write_one_node(_labels=_labels,
                                                 _aliases=_aliases,
                                                 _pagerank=_pagerank,
                                                 prev_node=prev_node,
                                                 output_file=output_file,
                                                 _descriptions=_descriptions,
                                                 add_all_text=add_text,
                                                 data_type=data_type,
                                                 instance_ofs=_instance_ofs,
                                                 qnode_statement_count=qnode_statement_count,
                                                 is_class=is_class,
                                                 wikitable_anchor_text=_wikitable_anchor_text,
                                                 wikipedia_anchor_text=_wikipedia_anchor_text,
                                                 abbreviated_name=_abbreviated_name,
                                                 redirect_text=_redirect_text,
                                                 text_embedding=_text_embedding,
                                                 graph_embeddings_complex=_graph_embeddings_complex,
                                                 graph_embeddings_transe=_graph_embeddings_transE,
                                                 ascii_labels=ascii_labels,
                                                 property_count=property_count,
                                                 class_count=class_count,
                                                 context=context,
                                                 is_human=is_human,
                                                 extra_aliases=_extra_aliases,
                                                 properties=_properties,
                                                 external_identifiers=_external_identifiers,
                                                 external_identifiers_pairs=_external_identifiers_pairs
                                                 )
        except Exception:
            print(traceback.print_exc())

        mapping_dict = ElasticsearchManager.create_mapping_es(languages=list(all_langs))
        open(mapping_file_path, 'w').write(json.dumps(mapping_dict))
        print('Done!')

    @staticmethod
    def _write_one_node(**kwargs):
        """
        inner function called by build_elasticsearch_file only
        :param kwargs:
        :return:
        """
        labels = kwargs["_labels"]
        aliases = kwargs["_aliases"]
        descriptions = kwargs["_descriptions"]
        _pagerank = kwargs["_pagerank"]
        prev_node = kwargs["prev_node"]
        output_file = kwargs["output_file"]
        add_all_text = kwargs['add_all_text']
        instance_ofs = kwargs['instance_ofs']
        data_type = kwargs['data_type']
        qnode_statement_count = kwargs['qnode_statement_count']
        is_class = kwargs['is_class']
        wikitable_anchor_text = kwargs['wikitable_anchor_text']
        wikipedia_anchor_text = kwargs['wikipedia_anchor_text']
        redirect_text = kwargs['redirect_text']
        text_embedding = kwargs['text_embedding']
        graph_embeddings_complex = kwargs['graph_embeddings_complex']
        graph_embeddings_transe = kwargs['graph_embeddings_transe']
        ascii_labels = list(kwargs['ascii_labels'])
        property_count = kwargs['property_count']
        class_count = kwargs['class_count']
        context = kwargs['context']
        is_human = kwargs['is_human']
        extra_aliases = kwargs['extra_aliases']
        properties = kwargs['properties']
        external_identifiers = kwargs['external_identifiers']
        external_identifiers_pairs = kwargs['external_identifiers_pairs']

        _labels = {}
        _aliases = {}
        _descriptions = {}
        _wikitable_anchor_text = {}
        _wikipedia_anchor_text = {}
        _abbreviated_name = {}
        _redirect_text = {}
        en_labels_aliases = set()

        for k in labels:
            _labels[k] = list(labels[k])
            if is_human:
                if k == 'en':
                    en_labels_aliases.update(_labels[k])

        for k in aliases:
            _aliases[k] = list(aliases[k])
            if is_human:
                if k == 'en':
                    en_labels_aliases.update(_aliases[k])

        for k in descriptions:
            _descriptions[k] = list(descriptions[k])

        for k in wikitable_anchor_text:
            _wikitable_anchor_text[k] = list(wikitable_anchor_text[k])

        for k in wikipedia_anchor_text:
            _wikipedia_anchor_text[k] = list(wikipedia_anchor_text[k])

        # for k in abbreviated_name:
        #     _abbreviated_name[k] = list(abbreviated_name[k])

        for k in redirect_text:
            _redirect_text[k] = list(redirect_text[k])

        abbreviated_names = set()
        for name in en_labels_aliases:
            abbreviated_names.update(ElasticsearchManager.generate_abbreviations(name))

        if len(_labels) > 0 or len(_aliases) > 0 or len(_descriptions) > 0:

            _ = {'id': prev_node,
                 'labels': _labels,
                 'aliases': _aliases,
                 'pagerank': _pagerank,
                 'descriptions': _descriptions,
                 'statements': qnode_statement_count,
                 'wikitable_anchor_text': _wikitable_anchor_text,
                 'wikipedia_anchor_text': _wikipedia_anchor_text,
                 'redirect_text': _redirect_text,
                 'qnode_alias': prev_node
                 }

            if add_all_text:
                _['all_text'] = ElasticsearchManager.create_all_text(_labels, aliases=_aliases,
                                                                     descriptions=_descriptions)

            if len(instance_ofs) > 0:
                _['instance_ofs'] = list(instance_ofs)
            if data_type is not None:
                _['data_type'] = data_type
            if is_class:
                _['is_class'] = 'true'
            if text_embedding:
                _['text_embedding'] = text_embedding
            if graph_embeddings_complex:
                _['graph_embedding_complex'] = graph_embeddings_complex
            if graph_embeddings_transe:
                _['graph_embeddings_transe'] = graph_embeddings_transe
            if len(ascii_labels) > 0:
                _['ascii_labels'] = ascii_labels
            if class_count:
                _['class_count'] = class_count
            if property_count:
                _['property_count'] = property_count
            if context:
                _['context'] = context
            if len(abbreviated_names) > 0:
                _['abbreviated_name'] = {'en': list(abbreviated_names)}
            if len(extra_aliases) > 0:
                _['extra_aliases'] = list(extra_aliases)
            if len(properties) > 0:
                _['properties'] = list(properties)
            if len(external_identifiers) > 0:
                _['external_identifiers'] = list(external_identifiers)
            if len(external_identifiers_pairs) > 0:
                _['external_identifiers_pairs'] = list(external_identifiers_pairs)
            output_file.write(json.dumps(_))

            output_file.write('\n')

    @staticmethod
    def remove_language_tag(label_str):
        return re.sub(r'@.*$', '', label_str).replace("'", "")

    @staticmethod
    def separate_language_text_tag(label_str):
        if len(label_str) == 0:
            return "", "en"
        if "@" in label_str:
            res = label_str.split("@")
            text_string = "@".join(res[:-1]).replace('"', "").replace("'", "")
            lang = res[-1].replace('"', '').replace("'", "")
        else:
            text_string = label_str.replace('"', "").replace("'", "")
            lang = "en"
        return text_string, lang

    @staticmethod
    def create_all_text(labels, aliases, descriptions):
        text = ''
        if 'en' in labels and labels['en']:
            text = text + '\n'.join(labels['en']) + '\n'
        if 'en' in aliases and aliases['en']:
            text = text + '\n'.join(aliases['en']) + '\n'
        if 'en' in descriptions and descriptions['en']:
            text = text + '\n'.join(descriptions['en']) + '\n'
        return text

    @staticmethod
    def to_float(input_str):
        try:
            return float(input_str)
        except Exception:
            return None

    @staticmethod
    def create_mapping_es(languages: List[str] = ['en']):
        fields_to_es_fields_mapping = {
            'abbreviated_name': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower'],
                'copy_to': [],
                'enabled': True
            },
            'aliases': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': ['all_labels.{}', 'all_labels_aliases'],
                'enabled': True
            },
            'all_labels': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower', 'keyword', 'ngram', 'trigram'],
                'copy_to': [],
                'enabled': True
            },
            'all_labels_aliases': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': [],
                'enabled': True
            },
            'anchor_text': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower', 'keyword', 'ngram'],
                'copy_to': [],
                'enabled': True
            },
            'ascii_labels': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': ['all_labels.en', 'all_labels_aliases'],
                'enabled': True
            },
            'extra_aliases': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': ['all_labels.en', 'all_labels_aliases'],
                'enabled': True
            },
            'external_identifiers': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': ['all_labels_aliases'],
                'enabled': True
            },
            'external_identifiers_pairs': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': [],
                'enabled': True
            },
            'class_count': {
                'enabled': False
            },
            'context': {
                'enabled': False
            },
            'data_type': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower'],
                'copy_to': [],
                'enabled': True
            },
            'descriptions': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower'],
                'copy_to': [],
                'enabled': True
            },
            'graph_embedding_complex': {
                'enabled': False
            }, 'graph_embeddings_transe': {
                'enabled': False
            },
            'id': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': [],
                'enabled': True
            },
            'instance_ofs': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower'],
                'copy_to': [],
                'enabled': True
            },
            'properties': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower'],
                'copy_to': [],
                'enabled': True
            },
            'is_class': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': [],
                'enabled': True
            },
            'labels': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': ['all_labels.{}', 'all_labels_aliases'],
                'enabled': True
            },
            'pagerank': {
                'field_type': 'float',
                'enabled': True
            },
            'property_count': {
                'enabled': False
            },
            'qnode_alias': {
                'field_type': 'text',
                'languages': [],
                'es_fields': ['keyword_lower', 'keyword'],
                'copy_to': ['all_labels.en', 'all_labels_aliases'],
                'enabled': True
            },
            'redirect_text': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower'],
                'copy_to': [],
                'enabled': True
            },
            'statements': {
                'field_type': 'integer',
                'enabled': True
            },
            'text_embedding': {
                'enabled': False
            },
            'wikipedia_anchor_text': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower'],
                'copy_to': ['anchor_text.{}'],
                'enabled': True
            },
            'wikitable_anchor_text': {
                'field_type': 'text',
                'languages': languages,
                'es_fields': ['keyword_lower'],
                'copy_to': ['anchor_text.{}'],
                'enabled': True
            }

        }

        properties_dict = {}

        for k in fields_to_es_fields_mapping:
            _ = fields_to_es_fields_mapping[k]
            properties_dict.update(ElasticsearchManager.create_mapping_part(field_name=k,
                                                                            field_type=_.get('field_type'),
                                                                            languages=_.get('languages', []),
                                                                            es_fields=_.get('es_fields', []),
                                                                            copy_to=_.get('copy_to', []),
                                                                            enabled=_.get('enabled')))

        settings = {
            "index": {
                "number_of_shards": 6,
                "analysis": {
                    "normalizer": {
                        "lowercase_normalizer": {
                            "filter": [
                                "lowercase"
                            ],
                            "type": "custom"
                        }
                    },
                    "analyzer": {
                        "edge_ngram_analyzer": {
                            "filter": [
                                "lowercase"
                            ],
                            "tokenizer": "edge_ngram_tokenizer"
                        },
                        "edge_ngram_search_analyzer": {
                            "tokenizer": "lowercase"
                        },
                        "trigram_analyzer": {
                            "tokenizer": "trigram_tokenizer",
                            "filter": [
                                "lowercase"
                            ]
                        }
                    },
                    "tokenizer": {
                        "edge_ngram_tokenizer": {
                            "token_chars": [
                                "letter"
                            ],
                            "min_gram": "2",
                            "type": "edge_ngram",
                            "max_gram": "20"
                        },
                        "trigram_tokenizer": {
                            "type": "ngram",
                            "min_gram": 3,
                            "max_gram": 4,
                            "token_chars": [
                                "letter"
                            ]
                        }
                    }
                }
            }
        }

        mapping_dict = {
            "mappings": {
                "properties": properties_dict
            },
            "settings": settings
        }
        return mapping_dict

    @staticmethod
    def load_elasticsearch_index(kgtk_jl_path, es_url, es_index, mapping_file_path=None, es_user=None,
                                 es_pass=None,
                                 batch_size=10000):
        """
         loads a jsonlines file to Elasticsearch index.

        Args:
            kgtk_jl_path: input json lines file, could be output of build_elasticsearch_index
            es_url:  Elasticsearch server url
            es_index: Elasticsearch index to be created/loaded
            mapping_file_path: mapping file for the index
            es_user: Elasticsearch user
            es_pass: Elasticsearch password
            batch_size: batch size to be loaded at once

        Returns: Nothing

        """

        # first create the index
        create_response = ElasticsearchManager.create_index(es_url, es_index, mapping_file_path, es_user, es_pass)
        print('create response: {}'.format(create_response.status_code))

        f = open(kgtk_jl_path)
        load_batch = []
        counter = 0

        for line in f:
            counter += 1

            each_res = line.replace('\n', '')
            if not each_res:
                continue
            json_x = json.loads(each_res)

            load_batch.append(json.dumps({"index": {"_id": json_x['id']}}))
            load_batch.append(line.replace('\n', ''))
            if len(load_batch) % batch_size == 0:
                counter += len(load_batch)
                print('done {} rows'.format(counter))
                response = None
                try:
                    response = ElasticsearchManager.load_index(es_url,
                                                               es_index,
                                                               '{}\n\n'.format('\n'.join(load_batch)),
                                                               es_user=es_user,
                                                               es_pass=es_pass)
                    if response.status_code >= 400:
                        print(response.text)
                except Exception:
                    print('Exception while loading a batch to es')
                    print(response.text)
                    print(response.status_code)
                load_batch = []

        if len(load_batch) > 0:

            response = ElasticsearchManager.load_index(es_url,
                                                       es_index,
                                                       '{}\n\n'.format('\n'.join(load_batch)),
                                                       es_user=es_user,
                                                       es_pass=es_pass)
            if response.status_code >= 400:
                print(response.text)
        print('Finished loading the elasticsearch index')

    @staticmethod
    def load_index(es_url, es_index, payload, es_user=None, es_pass=None):

        es_url_bulk = '{}/{}/_doc/_bulk'.format(es_url, es_index)

        headers = {
            'Content-Type': 'application/x-ndjson',
        }
        if es_user and es_pass:
            return requests.post(es_url_bulk, headers=headers, data=payload, auth=HTTPBasicAuth(es_user, es_pass))
        else:
            return requests.post(es_url_bulk, headers=headers, data=payload)

    @staticmethod
    def create_index(es_url, es_index, mapping_file_path, es_user=None, es_pass=None):
        es_url_index = '{}/{}'.format(es_url, es_index)
        # first check if index exists
        if es_user and es_pass:
            response = requests.get(es_url_index, auth=HTTPBasicAuth(es_user, es_pass))
        else:
            response = requests.get(es_url_index)

        if response.status_code == 200:
            print('Index: {} already exists...'.format(es_index))
        elif response.status_code // 100 == 4:
            if mapping_file_path is not None:
                # no need to create index if mapping file is not specified, it'll be created at load time
                mapping = json.load(open(mapping_file_path))
                if es_user and es_pass:
                    response = requests.put(es_url_index, auth=HTTPBasicAuth(es_user, es_pass), json=mapping)
                else:
                    response = requests.put(es_url_index, json=mapping)
                if response.text and "error" in json.loads(response.text):
                    pp = pprint.PrettyPrinter(indent=4)
                    pp.pprint(json.loads(response.text))
                    raise Exception("Creating new index failed! Please check the error response above!")

        else:
            print('An exception has occurred: ')
            print(response.text)
        return response

    @staticmethod
    def create_property_metadata_dict(property_file_path: str) -> dict:
        _ = {}
        f = gzip.open(property_file_path, 'rt')
        node1_idx = -1
        node2_idx = -1
        for line in f:
            vals = line.strip().split("\t")
            if 'node1' in vals and 'node2' in vals:
                node1_idx = vals.index('node1')
                node2_idx = vals.index('node2')
            else:
                _[vals[node1_idx]] = vals[node2_idx]

        return _

    @staticmethod
    def transliterate_label(label: str) -> str:
        ascii_label = ""
        try:
            ascii_label = unidecode(label)
        except Exception as e:
            print(e, f'input label: {label}')
        return ascii_label

    # noinspection PyTypeChecker
    @staticmethod
    def create_mapping_part(field_name: str,
                            field_type: str,
                            languages: List[str],
                            es_fields: List[str],
                            copy_to: List[str],
                            enabled: bool = True):
        if not enabled:
            return {
                field_name: {
                    "type": "object",
                    "enabled": False
                }
            }

        if field_type in ['integer', 'float']:
            return {
                field_name: {
                    "type": field_type
                }
            }

        _mapping = None
        if field_type == 'text':
            if len(languages) == 0:
                _mapping = {
                    field_name: ElasticsearchManager.create_es_fields_part(field_type, es_fields)
                }

                if len(copy_to) > 0:
                    _mapping[field_name]['copy_to'] = copy_to if isinstance(copy_to, list) else [copy_to]
            else:
                _mapping = {
                    field_name: {
                        'properties': {}
                    }
                }
                for language in languages:
                    _mapping[field_name]['properties'][language] = ElasticsearchManager.create_es_fields_part(
                        field_type, es_fields)
                    if len(copy_to) > 0:
                        _mapping[field_name]['properties'][language]['copy_to'] = [x.format(language)
                                                                                   for x
                                                                                   in copy_to]

        return _mapping

    # noinspection PyTypeChecker
    @staticmethod
    def create_es_fields_part(field_type: str, es_fields: List[str]):
        _mapping = {
            "type": field_type
        }

        if len(es_fields) > 0:
            _mapping['fields'] = {}
            for es_field in es_fields:
                if es_field == 'keyword':
                    _mapping['fields'][es_field] = {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                elif es_field == 'keyword_lower':
                    _mapping['fields'][es_field] = {
                        "type": "keyword",
                        "normalizer": "lowercase_normalizer"
                    }
                elif es_field == 'ngram':
                    _mapping['fields'][es_field] = {
                        "type": "text",
                        "analyzer": "edge_ngram_analyzer",
                        "search_analyzer": "edge_ngram_search_analyzer"
                    }
                elif es_field == "trigram":
                    _mapping['fields'][es_field] = {
                        "type": "text",
                        "analyzer": "trigram_analyzer"
                    }

        return _mapping

    @staticmethod
    def generate_abbreviations(name: str) -> List[str]:
        """
        Helper function to generate the abbreviation.
        Input: name_split: List of the words in a name
        Output: Abbreviated Name
        """
        name_split = name.split()
        abbreviated_names = set()

        abbr_label = ''
        for word in name_split[:-1]:
            abbr_label += word[0].upper() + '.' + ' '
        abbr_label += name_split[-1]

        abbreviated_names.add(abbr_label)

        if len(name_split) >= 2:
            abbr_label_end = name_split[-1] + ',' + ' '
            for word in name_split[:-1]:
                abbr_label_end += word[0].upper() + '.' + ' '

            abbreviated_names.add(abbr_label_end)

        for i in range(len(name_split) - 1):
            abbreviated_names.add(ElasticsearchManager._generate_abbr(name_split, i))

        return list(abbreviated_names)

    @staticmethod
    def _generate_abbr(name_split: List[str], word_index: int) -> str:
        abbr_label = ''
        for i in range(len(name_split) - 1):
            if i != word_index:
                abbr_label += name_split[i] + ' '
            else:
                abbr_label += name_split[i][0].upper() + '.' + ' '
        abbr_label += name_split[-1]
        return abbr_label

    @staticmethod
    def remove_text_inside_brackets(text, brackets="()"):
        count = [0] * (len(brackets) // 2)  # count open/close brackets
        saved_chars = []
        for character in text:
            for i, b in enumerate(brackets):
                if character == b:  # found bracket
                    kind, is_close = divmod(i, 2)
                    count[kind] += (-1) ** is_close  # `+1`: open, `-1`: close
                    if count[kind] < 0:  # unbalanced bracket
                        count[kind] = 0  # keep it
                    else:  # found bracket to remove
                        break
            else:  # character is not a [balanced] bracket
                if not any(count):  # outside brackets
                    saved_chars.append(character)
        return ''.join(saved_chars).strip()

    @staticmethod
    def parse_context_string(qnode, context_string: str) -> dict:

        context_dict = {
            qnode: []
        }

        p_v_dict = {}
        try:
            if context_string:
                prop_val_list = re.split(r'(?<!\\)\|', context_string)

                for prop_val in prop_val_list:
                    _type = prop_val[0]

                    assert _type in valid_context_types, f"Invalid context type :{_type} found, Qnode: {qnode}," \
                                                         f" property value string: {prop_val}"

                    values, property, item = ElasticsearchManager.parse_prop_val(qnode, prop_val)

                    if _type == 'd' and values.startswith("^"):
                        values = values.replace("^", "")

                    if item is None:
                        key = property
                        if key not in p_v_dict:
                            p_v_dict[key] = {
                                'p': property,
                                "t": _type,
                                'v': []
                            }
                        p_v_dict[key]['v'].append(values)

                    else:
                        key = f"{property}_{item}"
                        if key not in p_v_dict:
                            p_v_dict[key] = {
                                'p': property,
                                'i': item,
                                'v': [],
                                "t": _type
                            }
                        p_v_dict[key]['v'].append(values)
        except Exception as e:
            raise Exception(e)

        for k in p_v_dict:
            context_dict[qnode].append(p_v_dict[k])

        return context_dict

    @staticmethod
    def parse_prop_val(qnode: str, property_value_string: str) -> Tuple[str, str, Any]:
        line_started = False
        string_value = list()

        property_value_string = property_value_string[1:]

        for i, c in enumerate(property_value_string):
            if i == 0 and c == '"':  # start of the string value:
                line_started = True
            if line_started:
                string_value.append(c)
            if (i > 0 and c == '"' and property_value_string[i - 1] != "\\") \
                    or \
                    (i > 0 and c == '"' and property_value_string[i + 1] == ":" and property_value_string[
                        i + 2] == "P"):
                line_started = False
        string_val = "".join(string_value)
        length_remove = len(string_val) + 1  # ":" eg i"utc+01:00":P421:Q6655
        rem_vals = property_value_string[length_remove:].split(":")
        n = len(string_val)
        string_val = string_val[1: n - 1]

        property = rem_vals[0]

        if not property.startswith('P'):
            raise Exception(
                f"Unexpected format of the context string, Qnode: {qnode}, context string: {property_value_string}. "
                f"Property does not starts with 'P'")

        if len(rem_vals) == 2:
            item = rem_vals[1]
            if not item.startswith('Q'):
                raise Exception(
                    f"Unexpected format of the context string, Qnode: {qnode}, context string: {property_value_string}"
                    f". Item does not start with 'Q'")
            return string_val, property, item

        if len(rem_vals) == 1:
            return string_val, property, None

        if len(rem_vals) > 2:
            raise Exception(
                f"Unexpected format of the context string, Qnode: {qnode}, context string: {property_value_string}")



"""Convert edge file to html visualization
"""
import json
import math

from kgtk.cli_argparse import KGTKArgumentParser

from pathlib import Path
import sys
import typing

from kgtk.exceptions import KGTKException
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
from kgtk.kgtkformat import KgtkFormat
import re
from typing import List
from kgtk.utils.color_styles import color_style_dict, GRADIENT, CATEGORICAL


def parser():
    return {
        'help': 'Visualize API',
        'description': 'Use API to convert edge file to html visualization'
    }


kgtk_format = KgtkFormat()

compiled_hex_color_regex = re.compile(r"^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$")


class KgtkVisualize:
    def __init__(
            self,
            input_file: str = None,
            output_file: str = None,
            errors_to_stdout: bool = False,
            errors_to_stderr: bool = True,
            show_options: bool = False,
            verbose: bool = False,
            very_verbose: bool = False,
            node_file: str = None,
            direction: str = None,
            edge_label: bool = False,
            edge_color_column: str = None,
            edge_color_style: str = 'd3.interpolateRainbow',
            edge_color_default: str = '#000000',
            edge_width_column: str = None,
            edge_width_default: float = 1.0,
            edge_width_minimum: float = 1.0,
            edge_width_maximum: float = 5.0,
            edge_width_scale: str = None,
            edge_color_numbers: str = None,
            edge_color_hex: bool = False,
            node_color_column: str = None,
            node_color_style: str = 'd3.interpolateRainbow',
            node_color_default: str = '#000000',
            node_color_numbers: str = None,
            node_color_hex: bool = False,
            node_size_column: str = None,
            node_size_default: float = 2.0,
            node_size_minimum: float = 1.0,
            node_size_maximum: float = 5.0,
            node_size_scale: str = None,
            node_file_id: str = 'id',
            show_text_limit: int = 500,
            node_border_color: str = None,
            tooltip_column: str = None,
            show_text: str = None,
            show_blank_labels: bool = False,
            kwargs=None):
        if kwargs is None:
            kwargs = {'errors_to_stderr': True, 'show_options': False}

        if node_color_style not in color_style_dict:
            raise KGTKException(f'Unknown value: {node_color_style} for parameter: node_color_style. Please choose'
                                f' one of the following: {", ".join(list(color_style_dict))}')

        if edge_color_style not in color_style_dict:
            raise KGTKException(f'Unknown value: {edge_color_style} for parameter: edge_color_style. Please choose'
                                f' one of the following: {", ".join(list(color_style_dict))}')

        self.input_file = Path(input_file)
        self.output_file = Path(output_file)
        self.errors_to_stdout = errors_to_stdout
        self.errors_to_stderr = errors_to_stderr
        self.show_options = show_options
        self.verbose = verbose
        self.very_verbose = very_verbose
        self.node_file = node_file
        self.direction = direction
        self.edge_label = edge_label
        self.edge_color_column = edge_color_column
        self.edge_color_style = edge_color_style
        self.edge_color_default = edge_color_default
        self.edge_color_numbers = edge_color_numbers.lower() if edge_color_numbers is not None else edge_color_numbers
        self.edge_color_hex = edge_color_hex
        self.edge_width_column = edge_width_column
        self.edge_width_default = edge_width_default
        self.edge_width_minimum = edge_width_minimum
        self.edge_width_maximum = edge_width_maximum
        self.edge_width_scale = edge_width_scale
        self.node_color_column = node_color_column
        self.node_color_style = node_color_style
        self.node_color_default = node_color_default
        self.node_size_column = node_size_column
        self.node_size_default = node_size_default
        self.node_size_minimum = node_size_minimum
        self.node_size_maximum = node_size_maximum
        self.node_size_scale = node_size_scale
        self.node_file_id = node_file_id
        self.show_text_limit = show_text_limit
        self.node_border_color = node_border_color
        self.tooltip_column = tooltip_column
        self.show_text = show_text
        self.node_color_numbers = node_color_numbers.lower() if node_color_numbers is not None else node_color_numbers
        self.node_color_hex = node_color_hex
        self.show_blank_labels = show_blank_labels
        self.kwargs = kwargs

        self.input_kgtk_file: Path = KGTKArgumentParser.get_input_file(self.input_file)

        # Select where to send error messages, defaulting to stderr.
        self.error_file: typing.TextIO = sys.stdout if self.errors_to_stdout else sys.stderr
        # Build the option structures.
        self.reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(self.kwargs)
        self.value_options: KgtkValueOptions = KgtkValueOptions.from_dict(self.kwargs)

        # define the base for log
        self.base = 2.0

        self.edge_color_map = {}

        self.node_color_map = {}

        self.node_set = set()

        self.node_color_choice = 0
        self.edge_color_choice = 0

        if self.node_size_minimum == 0.0 and self.node_size_scale == 'log':
            raise ValueError("node size cannot be 0 when using log scale")
        if self.edge_width_minimum == 0 and self.edge_width_scale == 'log':
            raise ValueError("edge width cannot be 0 when using log scale")

    def execute(self) -> int:

        d = self.compute_visualization_graph()
        self.to_html(d)
        return 0

    def compute_visualization_graph(self):
        d = {}
        # Show the final option structures for debugging and documentation.
        try:
            edges, nodes_from_edge = self.process_edge_file()
            if self.node_file is not None:
                nodes = self.process_node_file(nodes_from_edge=nodes_from_edge)
            else:
                nodes = nodes_from_edge

            d['links'] = edges
            d['nodes'] = nodes
        except SystemExit as e:
            raise KGTKException(f"Exit requested: {e}")
        except Exception as e:
            raise KGTKException(str(e))
        return d

    def process_edge_file(self):
        # First create the KgtkReader.  It provides parameters used by the ID
        # column builder. Next, create the ID column builder, which provides a
        # possibly revised list of column names for the KgtkWriter.  Create
        # the KgtkWriter.  Last, process the data stream.
        # Open the input file.
        kr: KgtkReader = KgtkReader.open(self.input_kgtk_file,
                                         error_file=self.error_file,
                                         options=self.reader_options,
                                         value_options=self.value_options,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose,
                                         )

        nodes = set()
        nodes_from_edge_file = []
        edges = []

        node1_idx = kr.column_name_map['node1']
        node2_idx = kr.column_name_map['node2']

        if self.node_file is None and 'node1;label' in kr.column_name_map:
            node1_label_idx = kr.column_name_map['node1;label']
        else:
            node1_label_idx = kr.column_name_map['node1']

        if self.node_file is None and 'node2;label' in kr.column_name_map:
            node2_label_idx = kr.column_name_map['node2;label']
        else:
            node2_label_idx = kr.column_name_map['node2']

        if 'label;label' not in kr.column_name_map:
            label_label_idx = kr.column_name_map['label']
        else:
            label_label_idx = kr.column_name_map['label;label']

        for row in kr:
            if len(row) == len(kr.column_name_map):
                # collect nodes from the edge file,
                # incase node file is not present or a node is missing from the node file
                node1_label = row[node1_label_idx]
                node2_label = row[node2_label_idx]
                if '@' in node1_label:
                    # clean_node1_label, _, _ = kgtk_format.destringify(node1_label)
                    clean_node1_label = node1_label.split('@')[0]
                else:
                    clean_node1_label = node1_label
                nodes.add((row[node1_idx], clean_node1_label))

                if '@' in node2_label:
                    # clean_node2_label, _, _ = kgtk_format.destringify(node2_label)
                    clean_node2_label = node2_label.split('@')[0]
                else:
                    clean_node2_label = node2_label
                nodes.add((row[node2_idx], clean_node2_label))

                if '@' in row[label_label_idx]:
                    # _label_label, _, _ = kgtk_format.destringify(row[label_label_idx])
                    _label_label = row[label_label_idx].split('@')[0]
                else:
                    _label_label = row[label_label_idx]

                _edge_obj = {
                    'source': row[node1_idx],
                    'target': row[node2_idx],
                    'label': _label_label
                }

                if self.edge_width_column is not None:
                    _width_orig = KgtkVisualize.convert_string_float(row[kr.column_name_map[self.edge_width_column]])
                    if _width_orig is not None:
                        _edge_obj['width_orig'] = _width_orig
                    else:
                        _edge_obj['width_orig'] = 0.0
                else:
                    _edge_obj['width'] = self.edge_width_default

                if self.edge_color_column is not None:
                    _edge_color = row[kr.column_name_map[self.edge_color_column]].strip()

                    if self.edge_color_numbers is not None:
                        _edge_color = KgtkVisualize.convert_string_float(_edge_color)
                        if _edge_color is not None:
                            _edge_obj['orig_color'] = _edge_color
                        else:
                            _edge_obj['orig_color'] = 0.0
                    elif self.edge_color_hex:
                        if KgtkVisualize.is_valid_hex_color(_edge_color):
                            _edge_obj['color'] = _edge_color
                        else:
                            _edge_obj['color'] = self.edge_color_default
                    else:
                        _edge_obj['orig_color'] = _edge_color
                else:
                    _edge_obj['color'] = self.edge_color_default

                edges.append(_edge_obj)

        kr.close()

        if self.edge_width_column is not None:
            edges = self.calculate_size(edges,
                                        self.edge_width_scale,
                                        self.edge_width_default,
                                        self.edge_width_minimum,
                                        self.edge_width_maximum,
                                        'width_orig',
                                        'width')

        if self.edge_color_column is not None:
            edges = self.calculate_color(edges,
                                         self.edge_color_hex,
                                         self.edge_color_numbers,
                                         self.edge_color_style,
                                         self.edge_color_default,
                                         False)

        for ele in nodes:
            nodes_from_edge_file.append(
                {
                    'id': ele[0],
                    'label': ele[1],
                    'tooltip': ele[1],
                    'size': self.node_size_default,
                    'orig_size': self.node_size_default
                })

        return edges, nodes_from_edge_file

    @staticmethod
    def is_valid_hex_color(color_string: str) -> bool:

        if color_string is None or color_string.strip() == "":
            return False

        if re.search(compiled_hex_color_regex, color_string.strip()):
            return True
        else:
            return False

    @staticmethod
    def convert_string_float(a_string: str) -> float:
        if a_string is None or a_string.strip() == '':
            return None
        try:
            return float(a_string.strip())
        except ValueError:
            return None

    def process_node_file(self, nodes_from_edge):
        nodes = []
        if self.node_file is not None:

            kr_node: KgtkReader = KgtkReader.open(self.node_file,
                                                  error_file=self.error_file,
                                                  options=self.reader_options,
                                                  value_options=self.value_options,
                                                  verbose=self.verbose,
                                                  very_verbose=self.very_verbose,
                                                  mode=KgtkReaderMode.NONE,
                                                  )

            if self.node_file_id not in kr_node.column_name_map:
                raise ValueError("no id column in node file")

            for row in kr_node:
                if len(row) == len(kr_node.column_name_map):
                    _id = row[kr_node.column_name_map[self.node_file_id]]
                    temp = {'id': _id}
                    self.node_set.add(_id)

                    if 'x' in kr_node.column_name_map:
                        temp['fx'] = float(row[kr_node.column_name_map['x']])
                        temp['fy'] = float(row[kr_node.column_name_map['y']])

                    if 'label' in kr_node.column_name_map:
                        try:
                            _node_label = row[kr_node.column_name_map['label']]
                            if _node_label.strip() != '':
                                _node_label, _, _ = kgtk_format.destringify(_node_label)
                        except Exception as e:
                            print(e, file=self.error_file)
                            _node_label = row[kr_node.column_name_map['label']]

                        if _node_label != "" or self.show_blank_labels:
                            temp['label'] = _node_label
                        else:
                            temp['label'] = _id
                    else:
                        temp['label'] = _id

                    if self.tooltip_column is not None:
                        if str(row[kr_node.column_name_map[self.tooltip_column]]).strip() != '':
                            temp['tooltip'] = str(row[kr_node.column_name_map[self.tooltip_column]])
                        else:
                            temp['tooltip'] = temp['label']
                    else:
                        temp['tooltip'] = temp['label'] if temp['label'] != "" else _id

                    if self.node_color_column is not None:
                        _node_color = row[kr_node.column_name_map[self.node_color_column]]
                        if self.node_color_numbers is not None:
                            _node_color = KgtkVisualize.convert_string_float(_node_color)
                            if _node_color is not None:
                                temp['orig_color'] = _node_color
                            else:
                                temp['orig_color'] = 0.0
                        elif self.node_color_hex:
                            if KgtkVisualize.is_valid_hex_color(_node_color):
                                temp['color'] = _node_color
                            else:
                                temp['color'] = self.node_color_default
                        else:
                            temp['orig_color'] = _node_color
                    else:
                        temp['color'] = self.node_color_default

                    if self.node_size_column is not None:
                        _node_size = KgtkVisualize.convert_string_float(
                            row[kr_node.column_name_map[self.node_size_column]])
                        if _node_size is not None:
                            temp['orig_size'] = _node_size
                        else:
                            temp['orig_size'] = 0.0
                    else:
                        temp['size'] = self.node_size_default

                    nodes.append(temp)
            kr_node.close()

            for node_from_edge in nodes_from_edge:
                if node_from_edge['id'] not in self.node_set:
                    if self.node_color_numbers is not None:
                        node_from_edge['orig_color'] = 0.0
                    else:
                        node_from_edge['orig_color'] = self.node_color_default
                    nodes.append(node_from_edge)

            if self.node_color_column is not None:
                nodes = self.calculate_color(nodes,
                                             self.node_color_hex,
                                             self.node_color_numbers,
                                             self.node_color_style,
                                             self.node_color_default,
                                             True)

            if self.node_size_column is not None:
                self.calculate_size(nodes,
                                    self.node_size_scale,
                                    self.node_size_default,
                                    self.node_size_minimum,
                                    self.node_size_maximum,
                                    'orig_size',
                                    'size')

        return nodes

    def calculate_color(self,
                        nodes: List[dict],
                        node_color_hex: bool,
                        node_color_numbers: str,
                        node_color_style: str,
                        node_color_default: str,
                        process_nodes: bool) -> List[dict]:

        categorical_color_scale = False
        category_length = 1
        if node_color_hex:
            if process_nodes:
                self.node_color_choice = 0
            else:
                self.edge_color_choice = 0
            # all good, nothing to do here
            pass
        else:
            if process_nodes:
                self.node_color_choice = 1
            else:
                self.edge_color_choice = 1
            node_color_set = set()
            max_color = -1
            min_color = -1
            log_max_color = -1
            log_min_color = -1
            if node_color_numbers is not None:
                for x in nodes:
                    node_color_set.add(x['orig_color'])

                max_color = max(node_color_set)
                min_color = min(node_color_set)

                log_max_color = math.log(max_color, self.base) if max_color > 0.0 else -1.0
                log_min_color = math.log(min_color, self.base) if min_color > 0.0 else -1.0

                for i, color in enumerate(sorted(node_color_set)):
                    if process_nodes:
                        self.node_color_map[color] = i
                    else:
                        self.edge_color_map[color] = i
            for node in nodes:
                orig_color = node['orig_color']
                if node_color_numbers is not None:
                    if color_style_dict[node_color_style]['style'] == GRADIENT:
                        if node_color_numbers == 'linear':
                            node['color'] = (orig_color - min_color) / (max_color - min_color)
                        elif node_color_numbers == 'log':
                            if orig_color == 0.0 or log_max_color == log_min_color:
                                node['color'] = node_color_default
                            else:
                                node['color'] = (math.log(orig_color, self.base) - log_min_color) / (
                                        log_max_color - log_min_color)
                        else:
                            node['color'] = orig_color if orig_color != 0.0 else node_color_default

                if 'color' not in node:
                    if color_style_dict[node_color_style]['style'] == CATEGORICAL:
                        category_length = color_style_dict[node_color_style]['length']
                        categorical_color_scale = True
                    if process_nodes:
                        if orig_color not in self.node_color_map:
                            self.node_color_map[orig_color] = len(self.node_color_map)

                        if categorical_color_scale:
                            node['color'] = self.node_color_map[orig_color] % category_length
                        else:
                            node['color'] = self.node_color_map[orig_color]
                    else:
                        if orig_color not in self.edge_color_map:
                            self.edge_color_map[orig_color] = len(self.edge_color_map)
                        if categorical_color_scale:
                            node['color'] = self.edge_color_map[orig_color] % category_length
                        else:
                            node['color'] = self.edge_color_map[orig_color]
                del node['orig_color']
        return nodes

    def calculate_size(self,
                       nodes: List[dict],
                       node_size_scale: str,
                       node_size_default: str,
                       node_size_minimum: float,
                       node_size_maximum: float,
                       size_field: str,
                       output_field: str
                       ) -> List[dict]:
        node_size_list = [x[size_field] for x in nodes]
        max_size = max(node_size_list)
        min_size = min(node_size_list)
        log_max_size = math.log(max_size, self.base) if max_size > 0.0 else -1.0
        log_min_size = math.log(min_size, self.base) if min_size > 0.0 else -1.0
        for node in nodes:
            node_size = node[size_field]
            if node_size_scale == 'linear':
                node[output_field] = node_size_minimum + (node_size - min_size) * \
                                     (node_size_maximum - node_size_minimum) / \
                                     (max_size - min_size)
            elif node_size_scale == 'log':
                if node_size == 0.0 or log_min_size == log_max_size:
                    node[output_field] = node_size_default
                else:
                    node[output_field] = node_size_minimum + (math.log(node_size, self.base) - log_min_size) * (
                            node_size_maximum - node_size_minimum) / (log_max_size - log_min_size)
            else:
                node[output_field] = node_size if node_size > 0 else node_size_default
        return nodes

    def to_html(self, d):
        node_color_style = ''
        edge_color_style = ''

        if color_style_dict[self.node_color_style]['style'] == CATEGORICAL:
            node_color_style = f"{self.node_color_style}[node.color]"
        elif color_style_dict[self.node_color_style]['style'] == GRADIENT:
            node_color_style = f"d3.scaleSequential().domain([0, {len(self.node_color_map) - 1}]).interpolator({self.node_color_style})(node.color)"

        if color_style_dict[self.edge_color_style]['style'] == CATEGORICAL:
            edge_color_style = f"{self.edge_color_style}[link.color]"
        elif color_style_dict[self.edge_color_style]['style'] == GRADIENT:
            edge_color_style = f"d3.scaleSequential().domain([0, {len(self.edge_color_map) - 1}]).interpolator({self.edge_color_style})(link.color)"

        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(self.output_file)

        f = open(output_kgtk_file, 'w')
        f.write(f'''<head>
            <style> body {{ margin: 0; }} </style>
            <script src="https://cdn.jsdelivr.net/npm/d3-color@3"></script>
            <script src="https://cdn.jsdelivr.net/npm/d3-interpolate@3"></script>
            <script src="https://cdn.jsdelivr.net/npm/d3-scale-chromatic@3"></script>
            <script src="https://cdn.jsdelivr.net/npm/d3-scale@4"></script>
            <script src="https://unpkg.com/force-graph"></script>
            <!--<script src="../../dist/force-graph.js"></script>-->
            </head>
            <body>
            <div id="graph"></div>
            <script>        
               const j = ''')
        f.write(json.dumps(d, indent=4))
        f.write('''
              const Graph = ForceGraph()
              (document.getElementById('graph'))
                .graphData(j)
                .nodeId('id')
                .nodeLabel('tooltip')
                .nodeVal('size')''')

        if self.node_color_choice == 0:
            f.write('''
                        .nodeAutoColorBy('group')
                        .linkWidth((link) => link.width)''')
            node_text_format = '(node.color)'
        else:
            f.write(f'''
                    .nodeColor((node) => node.color[0] == "#" ? node.color : {node_color_style})
                    .linkWidth((link) => link.width)''')
            node_text_format = node_color_style

        if self.node_border_color is not None:
            node_text_format = "'" + self.node_border_color + "'"
        if self.direction == 'arrow':
            f.write('''
                  .linkDirectionalArrowLength(6)
                  .linkDirectionalArrowRelPos(1)''')
        elif self.direction == 'particle':
            f.write('''      .linkDirectionalParticles(2)
                ''')
        if self.show_text is not None and self.show_text_limit > len(d['nodes']):

            if self.show_text == 'center':
                y_move = 0
            else:
                y_move = 10

            f.write('''
                            .nodeCanvasObject((node, ctx, globalScale) => {
                      const label = node.label;
                      const fontSize = 12/globalScale;
                      ctx.font = `${fontSize}px Sans-Serif`;
                      const textWidth = ctx.measureText(label).width;
                      const bckgDimensions = [textWidth, fontSize].map(n => n + fontSize * 0.2); // some padding
                      ''')

            f.write(f'''
                      ctx.fillStyle = 'rgba(255, 255, 255, 0.8)';
                      ctx.fillRect(node.x - bckgDimensions[0] / 2, node.y - {y_move} - 
                      bckgDimensions[1] / 2, ...bckgDimensions);
                      ctx.textAlign = 'center';
                      ctx.textBaseline = 'middle';
                      ''')

            if node_text_format != '':
                f.write(f'''
                        ctx.fillStyle = {node_text_format};
                        ''')
            f.write(f'''
                      ctx.fillText(label, node.x, node.y - {y_move});
                      ''')

            f.write('''
                      ctx.beginPath(); ctx.arc(node.x, node.y, node.size, 0, 2 * Math.PI, false);  ctx.fill();
                      node.__bckgDimensions = bckgDimensions; // to re-use in nodePointerAreaPaint
                      })''')
        if self.edge_color_choice == 0:
            f.write('''        .linkColor((link) => link.color)''')
        else:
            f.write(f'''        
                .linkColor((link) => link.color[0] == "#" ? link.color : {edge_color_style})''')

        if self.edge_label:
            f.write('''                        .linkCanvasObjectMode(() => 'after')
                .linkCanvasObject((link, ctx) => {
                  const MAX_FONT_SIZE = 4;
                  const LABEL_NODE_MARGIN = Graph.nodeRelSize() * 1.5;
                  const start = link.source;
                  const end = link.target;
                  // ignore unbound links
                  if (typeof start !== 'object' || typeof end !== 'object') return;
                  // calculate label positioning
                  const textPos = Object.assign(...['x', 'y'].map(c => ({
                    [c]: start[c] + (end[c] - start[c]) / 2 // calc middle point
                  })));
                  const relLink = { x: end.x - start.x, y: end.y - start.y };
                  const maxTextLength = Math.sqrt(Math.pow(relLink.x, 2) + Math.pow(relLink.y, 2)) - LABEL_NODE_MARGIN * 2;
                  let textAngle = Math.atan2(relLink.y, relLink.x);
                  // maintain label vertical orientation for legibility
                  if (textAngle > Math.PI / 2) textAngle = -(Math.PI - textAngle);
                  if (textAngle < -Math.PI / 2) textAngle = -(-Math.PI - textAngle);
                  const label = `${link.label}`;
                  // estimate fontSize to fit in link length
                  const color = `rgba(${link.color}, 0.8)`;
                  ctx.font = '1px Sans-Serif';
                  const fontSize = Math.min(MAX_FONT_SIZE, maxTextLength / ctx.measureText(label).width);
                  ctx.font = `${fontSize}px Sans-Serif`;
                  const textWidth = ctx.measureText(label).width;
                  const bckgDimensions = [textWidth, fontSize].map(n => n + fontSize * 0.2); // some padding
                  // draw text label (with background rect)
                  ctx.save();
                  ctx.translate(textPos.x, textPos.y);
                  ctx.rotate(textAngle);
                  ctx.fillStyle = 'rgba(255, 255, 255)';
                  ctx.fillRect(- bckgDimensions[0] / 2, - bckgDimensions[1] / 2, ...bckgDimensions);
                  ctx.textAlign = 'center';
                  ctx.textBaseline = 'middle';
                  ctx.fillStyle = 'darkgrey';
                  ctx.fillText(label, 0, 0);
                  ctx.restore();
                });
                ''')
        f.write('''  </script>
            </body>''')
        f.close()



"""
Import CSV file in Graph-tool.

Note:  the log file wasn't converted to the new filename parsing API.

Note: ID generation is fixed.  The standard ID generator should be used instead.
"""
from argparse import Namespace
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import a TSV or CSV file in Graph-tool, optionally generating various statistics ' +
        '(pagerank, in-degrees, out-degrees, degree distribution, and hits).  Two different output ' +
        'files are generated: a log file, containing a combination of TSV data nd plain text records, ' +
        'and a KGTK file. Summary statistics are directed to ' +
        'the text log file, while detailed statistics (in-degrees, out-degrees, pagerank) are directed ' +
        'to the KGTK output file.' +
        '\nBy default, in-degrees, out-degrees, and pageranks will be computed and output. ' +
        'HITS properties will be computed an doutput for directed graphs, but not for undirected ones.'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True, optional=False)
    parser.add_output_file()

    parser.add_argument('--undirected', dest="undirected",
                        help='Is the graph undirected? If false, then the graph is ' +
                        ' treated as (node1)->(node2).  If true, then the graph is ' +
                        ' treated as (node1)<->(node2). ' +
                        '\nAlso, HITS will not be computed on undirected graphs. ' +
                        '\n(default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=False, metavar='True|False')

    parser.add_argument('--compute-pagerank', dest='compute_pagerank',
                        help='Whether or not to compute the PageRank property. ' +
                        '\nNote: --undirected improves the pagerank calculation. ' +
                        'If you want both pagerank and in/out-degrees, you should make two runs. ' +
                        '\n(default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')
                        
    parser.add_argument('--compute-hits', dest='compute_hits',
                        help='Whether or not to compute the HITS properties. ' +
                        '\nNote: --undirected disables HITS calculation. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--output-statistics-only', dest='output_statistics_only',
                        help='If this option is set, write only the statistics edges to the primary output file. ' +
                        'Else, write both the statistics and the original graph. (default=%(default)s',
                        type=optional_bool, nargs='?', const=True, default=False, metavar='True|False')

    parser.add_argument('--output-degrees', dest='output_degrees',
                        help='Whether or not to write degree edges to the primary output file. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--output-pagerank', dest='output_pagerank',
                        help='Whether or not to write pagerank edges to the primary output file. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--output-hits', dest='output_hits',
                        help='Whether or not to write HITS edges to the primary output file. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--log-file', action='store', type=str, dest='log_file',
                        help='Summary file for the global statistics of the graph.', default='./summary.txt')

    parser.add_argument('--log-top-relations', dest='log_top_relations',
                        help='Whether or not to compute top relations and output them to the log file. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--log-degrees-histogram', dest='log_degrees_histogram',
                        help='Whether or not to compute degree distribution and output it to the log file. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--log-top-pageranks', dest='log_top_pageranks',
                        help='Whether or not to output PageRank centrality top-n to the log file. ' +
                        '\n(default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')
                        
    parser.add_argument('--log-top-hits', dest='log_top_hits',
                        help='Whether or not to output the top-n HITS to the log file. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=True, metavar='True|False')

    parser.add_argument('--log-top-n', action='store', dest='top_n',
                        default=5, type=int,
                        help='Number of top centrality nodes to write to the log file. (default=%(default)d)')

    parser.add_argument('--vertex-in-degree-property', action='store', dest='vertex_in_degree',
                        default='vertex_in_degree',
                        help='Label for edge: vertex in degree property. ' +
                        '\nNote: If --undirected is True, then the in-degree will be 0. ' +
                        '\n(default=%(default)s')

    parser.add_argument('--vertex-out-degree-property', action='store', dest='vertex_out_degree',
                        default='vertex_out_degree',
                        help='Label for edge: vertex out degree property. ' +
                        '\nNote: if --undirected is True, the the out-degree will be the sum of ' +
                        'the values that would have been calculated for in-degree and -out-degree ' +
                        ' if --undirected were False. ' +
                        '\n(default=%(default)s)')

    parser.add_argument('--page-rank-property', action='store', dest='vertex_pagerank',
                        default='vertex_pagerank',
                        help='Label for pank rank property. (default=%(default)s)')
    
    parser.add_argument('--vertex-hits-authority-property', action='store', dest='vertex_auth',
                        default='vertex_auth',
                        help='Label for edge: vertext hits authority. (default=%(default)s)')
    
    parser.add_argument('--vertex-hits-hubs-property', action='store', dest='vertex_hubs',
                        default='vertex_hubs',
                        help='Label for edge: vertex hits hubs. (default=%(default)s)')

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        
        undirected: bool,
        compute_pagerank: bool,
        compute_hits: bool,

        output_statistics_only: bool,
        output_degrees: bool,
        output_pagerank: bool,
        output_hits: bool,
        
        log_file: str,
        log_degrees_histogram: bool,
        log_top_relations: bool,
        log_top_pageranks: bool,
        log_top_hits: bool,
        top_n: int,

        vertex_in_degree: str,
        vertex_out_degree: str,
        vertex_pagerank: str,
        vertex_auth: str,
        vertex_hubs: str,

        errors_to_stdout: bool,
        errors_to_stderr: bool,
        show_options: bool,
        verbose: bool,
        very_verbose: bool,

        **kwargs, # Whatever KgtkFileOptions and KgtkValueOptions want.
        ):
    # import modules locally
    from pathlib import Path
    import sys

    from graph_tool import centrality
    from kgtk.exceptions import KGTKException
    import kgtk.gt.analysis_utils as gtanalysis
    from kgtk.gt.gt_load import load_graph_from_kgtk
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    try:

        # Select where to send error messages, defaulting to stderr.
        error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

        # Build the option structures.
        reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
        value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

        input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

        # hardcoded values useful for the script. Perhaps some of them should be exposed as arguments later
        directions = ['in', 'out', 'total']
        id_col = 'name'
        output_columns = ["node1", "label", "node2", "id"]

        if verbose:
            print('loading the KGTK input file...\n', file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         error_file=error_file,
                                         options=reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
                                         )
        sub: int = kr.get_node1_column_index()
        if sub < 0:
            print("Missing node1 (subject) column.", file=error_file, flush=True)
        pred: int = kr.get_label_column_index()
        if pred < 0:
            print("Missing label (predicate) column.", file=error_file, flush=True)
        obj: int = kr.get_node2_column_index()
        if obj < 0:
            print("Missing node2 (object) column", file=error_file, flush=True)
        if sub < 0 or pred < 0 or obj < 0:
            kr.close()
            raise KGTKException("Exiting due to missing columns.")

        predicate: str = kr.column_names[pred]

        G2 = load_graph_from_kgtk(kr, directed=not undirected, ecols=(sub, obj), verbose=verbose, out=error_file)
        if verbose:
            print('graph loaded! It has %d nodes and %d edges.' % (G2.num_vertices(), G2.num_edges()), file=error_file, flush=True)
        kr.close()

        if compute_pagerank:
            if verbose:
                print('Computing pagerank.', file=error_file, flush=True)
            v_pr = G2.new_vertex_property('float')
            centrality.pagerank(G2, prop=v_pr)
            G2.properties[('v', vertex_pagerank)] = v_pr

        if compute_hits and not undirected:
            if verbose:
                print('Computing HITS.', file=error_file, flush=True)
            hits_eig, G2.vp[vertex_hubs], G2.vp[vertex_auth] = gtanalysis.compute_hits(G2)

        if verbose:
            print('Opening the output file: %s' % repr(str(output_kgtk_file)), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(output_columns,
                                     output_kgtk_file,
                                     mode=KgtkWriter.Mode.EDGE,
                                     require_all_columns=True,
                                     prohibit_extra_columns=True,
                                     fill_missing_columns=False,
                                     verbose=verbose,
                                     very_verbose=very_verbose)

        if not output_statistics_only:
            if verbose:
                print('Copying the input edges to the output.', file=error_file, flush=True)
            id_count = 0
            for e in G2.edges():
                sid, oid = e
                lbl = G2.ep[predicate][e]
                kw.write([G2.vp[id_col][sid], lbl, G2.vp[id_col][oid], '{}-{}-{}'.format(G2.vp[id_col][sid], lbl, id_count)])
                id_count += 1

        if output_degrees or output_pagerank or output_hits:
            if verbose:
                print('Outputting vertex degrees and/or properties.', file=error_file, flush=True)
            id_count = 0
            for v in G2.vertices():
                v_id = G2.vp[id_col][v]
                if output_degrees:
                    kw.write([v_id, vertex_in_degree, str(v.in_degree()), '{}-{}-{}'.format(v_id, vertex_in_degree, id_count)])
                    id_count += 1
                    kw.write([v_id, vertex_out_degree, str(v.out_degree()), '{}-{}-{}'.format(v_id, vertex_out_degree, id_count)])
                    id_count += 1

                if output_pagerank:
                    if vertex_pagerank in G2.vp:
                        kw.write([v_id, vertex_pagerank, str(G2.vp[vertex_pagerank][v]), '{}-{}-{}'.format(v_id, vertex_pagerank, id_count)])
                        id_count += 1

                if output_hits:
                    if vertex_auth in G2.vp:
                        kw.write([v_id, vertex_auth, str(G2.vp[vertex_auth][v]), '{}-{}-{}'.format(v_id, vertex_auth, id_count)])
                        id_count += 1
                    if vertex_hubs in G2.vp:
                        kw.write([v_id, vertex_hubs, str(G2.vp[vertex_hubs][v]), '{}-{}-{}'.format(v_id, vertex_hubs, id_count)])
                        id_count += 1


        kw.close()

        if verbose:
            print('Writing the summary file.', file=error_file, flush=True)
        with open(log_file, 'w') as writer:
            writer.write('graph loaded! It has %d nodes and %d edges\n' % (G2.num_vertices(), G2.num_edges()))
            if log_top_relations:
                writer.write('\n*** Top relations:\n')
                for rel, freq in gtanalysis.get_topN_relations(G2, pred_property=predicate):
                    writer.write('%s\t%d\n' % (rel, freq))

            if log_degrees_histogram:
                writer.write('\n*** Degrees:\n')
                for direction in directions:
                    degree_data = gtanalysis.compute_node_degree_hist(G2, direction)
                    max_degree = len(degree_data) - 1
                    mean_degree, std_degree = gtanalysis.compute_avg_node_degree(G2, direction)
                    writer.write(
                        '%s degree stats: mean=%f, std=%f, max=%d\n' % (direction, mean_degree, std_degree, max_degree))

            if log_top_pageranks and compute_pagerank:
                writer.write('\n*** PageRank\n')
                writer.write('Max pageranks\n')
                result = gtanalysis.get_topn_indices(G2, vertex_pagerank, top_n, id_col)
                for n_id, n_label, pr in result:
                    writer.write('%s\t%s\t%f\n' % (n_id, n_label, pr))

            if log_top_hits and compute_hits and not undirected:
                writer.write('\n*** HITS\n')
                writer.write('HITS hubs\n')
                main_hubs = gtanalysis.get_topn_indices(G2, vertex_hubs, top_n, id_col)
                for n_id, n_label, hubness in main_hubs:
                    writer.write('%s\t%s\t%f\n' % (n_id, n_label, hubness))
                writer.write('HITS auth\n')
                main_auth = gtanalysis.get_topn_indices(G2, vertex_auth, top_n, id_col)
                for n_id, n_label, authority in main_auth:
                    writer.write('%s\t%s\t%f\n' % (n_id, n_label, authority))

    except Exception as e:
        raise KGTKException('Error: ' + str(e))



"""
Make a side copy of the input stream.

--mode=NONE is the degault
"""
from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Make a side copy of the input/output stream..',
        'description': 'Copy the primary input to the primary output, making a copy to a specified file. This can be used to make a copy of a pipe\'s data.' +
        '\n\nThis command defaults to --mode=NONE so it will work with TSV files that do not follow KGTK column naming conventions.'
        
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file()
    parser.add_output_file()

    parser.add_output_file(who="The tee output file", dest="into_file", options=["--into-file"], metavar="INTO_FILE",
                           optional=False, allow_list=False, allow_stdout=False, default_stdout=False)

    parser.add_argument(      "--enable", dest="enable", help="When True, enable copying the data stream to the --into-file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        into_file: KGTKFiles,

        enable: bool,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import kgtk_exception_auto_handler, KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    into_kgtk_file: Path = KGTKArgumentParser.get_output_file(into_file, who="The tee output file", default_stdout=False)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        print("--to-file=%s" % str(into_kgtk_file), file=error_file)
        print("--enable=%s" % str(enable), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:

        if verbose:
            print("Opening the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        if verbose:
            print("Opening the output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        tkw: typing.Optional[KgtkWriter] = None
        if enable:
            if verbose:
                print("Opening the tee output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
            tkw = KgtkWriter.open(kr.column_names,
                                  into_kgtk_file,
                                  mode=KgtkWriter.Mode[kr.mode.name],
                                  verbose=verbose,
                                  very_verbose=very_verbose)

        input_line_count: int = 0
        row: typing.List[str]
        for row in kr:
            input_line_count += 1
            kw.write(row)
            if tkw is not None:
                tkw.write(row)

        if verbose:
            print("Processed %d rows." % (input_line_count), file=error_file, flush=True)

        kw.close()
        if tkw is not None:
            tkw.close()

        return 0

    except Exception as e:
        kgtk_exception_auto_handler(e)
        return 1



"""
SQLStore and Kypher utilities.
"""

import math
from   odictliteral import odict
import pprint
import re

import sh

from   kgtk.exceptions import KGTKException

pp = pprint.PrettyPrinter(indent=4)


### Utilities

def listify(x):
    return (hasattr(x, '__iter__') and not isinstance(x, str) and list(x)) or (x and [x]) or []

# TO DO: I am sure some form of this already exists somewhere in Craig's code

def open_to_read(file, mode='rt'):
    """Version of 'open' that is smart about different types of compressed files
    and file-like objects that are already open to read.  'mode' has to be a
    valid read mode such as 'r', 'rb' or 'rt'.
    """
    assert mode in ('r', 'rb', 'rt'), 'illegal read mode'
    enc = 't' in mode and 'utf8' or None
    if isinstance(file, str) and file.endswith('.gz'):
        import gzip
        return gzip.open(file, mode, encoding=enc)
    elif isinstance(file, str) and file.endswith('.bz2'):
        import bz2
        return bz2.open(file, mode, encoding=enc)
    elif isinstance(file, str) and file.endswith('.xz'):
        import lzma
        return lzma.open(file, mode, encoding=enc)
    elif hasattr(file, 'read'):
        return file
    else:
        return open(file, mode)

def open_to_write(file, mode='wt'):
    """Version of 'open' that is smart about different types of compressed files
    and file-like objects that are already open to write.  'mode' has to be a
    valid write mode such as 'w', 'wb' or 'wt'.
    """
    assert mode in ('w', 'wb', 'wt'), 'illegal write mode'
    enc = 't' in mode and 'utf8' or None
    if isinstance(file, str) and file.endswith('.gz'):
        import gzip
        return gzip.open(file, mode, encoding=enc)
    elif isinstance(file, str) and file.endswith('.bz2'):
        import bz2
        return bz2.open(file, mode, encoding=enc)
    elif isinstance(file, str) and file.endswith('.xz'):
        import lzma
        return lzma.open(file, mode, encoding=enc)
    elif hasattr(file, 'write'):
        return file
    else:
        return open(file, mode)

def get_cat_command(file, _piped=False):
    """Return a cat-like sh-command to copy the possibly compressed 'file' to stdout.
    """
    # This works around some cross-platform issues with similar functionality in zconcat.
    if file.endswith('.gz'):
        return sh.gunzip.bake('-c', file, _piped=_piped)
    elif file.endswith('.bz2'):
        return sh.bunzip2.bake('-c', file, _piped=_piped)
    elif file.endswith('.xz'):
        return sh.unxz.bake('-c', file, _piped=_piped)
    else:
        return sh.cat.bake(file, _piped=_piped)

def format_memory_size(bytes):
    """Return a human-readable formatting of 'bytes' using powers of 1000.
    """
    units = ('Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    factor = 1000
    if bytes < factor:
        return '%d %s' % (bytes, units[0])
    else:
        scale = min(math.floor(math.log(bytes, factor)), len(units)-1)
        return "%.2f %s" % (bytes / pow(factor, scale), units[scale])

def parse_memory_size(size):
    """Parse a human-readable 'size' spec into the corresponding number (of bytes).
    If 'size' is already an integer, return it as is.
    """
    if isinstance(size, str):
        try:
            units = 'BKMGTPEZY'
            m = re.match(f'^\s*([0-9.]+)\s*([{units}]?I?B?\s*)$', size.upper())
            number, unit = m.group(1), m.group(2)
            # K and KiB is 1024, KB is 1000 (and similar for other units):
            factor = 1000 if 'B' in unit and 'I' not in unit else 1024
            scale = units.find(unit) if len(unit) <= 1 else units.find(unit[0])
            # only coerce to float if we have to to avoid precision issues:
            number = float(number) if '.' in number else int(number)
            size = number * pow(factor, scale)
        except:
            raise KGTKException(f'illegal size spec: {size}')
    return int(size)

def sql_quote_ident(ident, quote='"'):
    # - standard SQL quoting for identifiers such as table and column names is via double quotes
    # - double quotes within identifiers can be escaped via two double quotes
    # - sqlite also supports MySQL's backtick syntax and SQLServer's [] syntax
    return quote + ident.replace(quote, quote+quote) + quote


class sdict(odict):
    """Ordered schema dict that supports property access of its elements.
    """
    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __repr__(self):
        """Create an eval-able repr that will recreate 'self' identically."""
        if len(self) == 0:
            return "sdict()"
        else:
            return "sdict[%s]" % (", ".join("%s: %s" % (repr(k),repr(v)) for k,v in self.items()),)



"""
Constants and helpers for the KGTK file format.

"""

import ast
import datetime as dt
from enum import Enum, unique
import sys
import typing

class KgtkFormat:
    COLUMN_SEPARATOR: str = "\t"
    COMMENT_INDICATOR: str = "#"
    LIST_SEPARATOR: str = "|"

    # This value will be used to seperate fields when building a composit
    # sorting/grouping key.  It will work so long as \0 does not appear as
    # part of any of the key fields.
    KEY_FIELD_SEPARATOR: str = '\0'

    # These are the required columns in an edge file:
    NODE1_COLUMN_NAMES: typing.List[str] = ["node1", "from", "subject"]
    NODE2_COLUMN_NAMES: typing.List[str] = ["node2", "to", "object"]
    LABEL_COLUMN_NAMES: typing.List[str] = ["label", "predicate", "relation", "relationship"]

    # There is only one required column in a node file:
    ID_COLUMN_NAMES: typing.List[str] = ["id", "ID"]

    # These are the canonical names:
    NODE1: str = "node1"
    LABEL: str = "label"
    NODE2: str = "node2"
    ID: str = "id"

    KGTK_NAMESPACE: str = "kgtk:"

    @unique
    class DataType(Enum):
        EMPTY = 0
        LIST = 1
        NUMBER = 2
        QUANTITY = 3
        STRING = 4
        LANGUAGE_QUALIFIED_STRING = 5
        LOCATION_COORDINATES = 6
        DATE_AND_TIMES = 7
        EXTENSION = 8
        BOOLEAN = 9
        SYMBOL = 10

        def lower(self)->str:
            return self.name.lower()

        @classmethod
        def choices(cls)->typing.List[str]:
            results: typing.List[str] = [ ]
            name: str
            for name in cls.__members__.keys():
                results.append(name.lower())
            return results

    STRING_SIGIL: str = '"'
    LANGUAGE_QUALIFIED_STRING_SIGIL: str = "'"
    DATE_AND_TIMES_SIGIL: str = "^"
    LOCATION_COORDINATES_SIGIL: str = "@"

    TRUE_SYMBOL: str = "True"
    FALSE_SYMBOL: str = "False"

    @classmethod
    def to_boolean(cls, b: bool)->str:
        return cls.TRUE_SYMBOL if b else cls.FALSE_SYMBOL

    @classmethod
    def from_boolean(cls, item: str)->bool:
        return item == cls.TRUE_SYMBOL

    stringify_translate = str.maketrans({
        "\a": "\\a", # alarm (bell) - ASCII <BEL>
        "\b": "\\b", # backspace - ASCII <BS>
        "\f": "\\f", # formfeed - ASCII <FF>
        "\n": "\\n", # newline (linefeed) - ASCII <LF>
        "\r": "\\r", # carriage return -- ASCII <CR>
        "\t": "\\t", # horizontal tab - ASCII <TAB>
        "\v": "\\v", # vertical tab - ASCII <VT>
        "\\" : "\\\\", # backslash  - (\)
        "'": "\\'", # single quote - (')
        '"': '\\"', # double quote - (")
        LIST_SEPARATOR: "\\" + LIST_SEPARATOR, # vertical bar (pipe) - (|)
    })

    @classmethod
    def stringify(cls,
                  s: str,
                  language: str = "",
                  language_suffix: str = "",
    )->str:
        """Convert an internal string into a KGTK format string.  The internal string
        shouldn't have any "\" escaped characters.  For example, <TAB>
        characters should be ASCII tabs, not the two-character sequence
        backslash t ("\t").

        If a language code is provided, then a KGTK language qualified string
        is produced.  Otherwise, an ordinary KGTK string is produced.  In both
        cases, internal single and doublle quotes are both protected by
        backslashes.

        TODO: Should we octal encode <NUL>, <DEL> and any remaining ASCII control characters?
        """
        if len(language) == 0:
            return cls.STRING_SIGIL + s.translate(KgtkFormat.stringify_translate) + cls.STRING_SIGIL
        else:
            return cls.LANGUAGE_QUALIFIED_STRING_SIGIL + s.translate(KgtkFormat.stringify_translate) + "'@" + language + language_suffix

    @classmethod
    def unstringify(cls, s: str, unescape_pipe: bool = True)->str:
        """Convert a KGTK formatted string into an internal string.  The language
        code and suffix are not returned.
        """
        if s.startswith(cls.LANGUAGE_QUALIFIED_STRING_SIGIL):
            language: str
            s, language = s.rsplit("@", 1)
        if unescape_pipe:
            s = s.replace('\\|', '|')
        return ast.literal_eval(s)

    @classmethod
    def destringify(cls, s: str)->typing.Tuple[str, str, str]:
        """Convert a KGTK formatted string into an internal string.  The language
        code and suffix are returned.  Language and language_suffix are returned as empty
        strings when they are not present.

        For the moment, we'll assume that the language and language suffix don't
        conatain any escape sequences.
        """
        language: str = ""
        language_suffix: str = ""
        if s.startswith(cls.LANGUAGE_QUALIFIED_STRING_SIGIL):
            s, language = s.rsplit("@", 1)
            if "-" in language:
                language, language_suffix = language.split("-", 1)
                language_suffix = "-" + language_suffix
        s = s.replace('\\|', '|')
        return (ast.literal_eval(s), language, language_suffix)
    
    @classmethod
    def year(cls, year: typing.Union[int, str])->str:
        """Convert a year (passed as an integer or string) to a KGTK value for the year as a period.
        """
        yearstr: str
        if isinstance(year, int):
            yearstr = str(year)
        else:
            yearstr = year
        return cls.DATE_AND_TIMES_SIGIL + yearstr + "-01-01T00:00:00/9"

    @classmethod
    def year_month(cls,
                   year: typing.Union[int, str],
                   month: typing.Union[int, str])->str:
        """Convert a year and month (passed as integers or strings) to a KGTK value for the period.
        """
        yearstr: str
        if isinstance(year, int):
            yearstr = str(year)
        else:
            yearstr = year

        monthstr: str
        if isinstance(month, int):
            monthstr = str(month)
        else:
            monthstr = month
        if len(monthstr) == 1:
            monthstr = "0" + monthstr
        return cls.DATE_AND_TIMES_SIGIL + yearstr + "-" + monthstr + "-01T00:00:00/10"

    @classmethod
    def year_month_day(cls,
                       year: typing.Union[int, str],
                       month: typing.Union[int, str],
                       day: typing.Union[int, str],
                       )->str:
        """Convert a year, month, and day (passed as integers or strings) to a KGTK value for the period.
        """
        yearstr: str
        if isinstance(year, int):
            yearstr = str(year)
        else:
            yearstr = year

        monthstr: str
        if isinstance(month, int):
            monthstr = str(month)
        else:
            monthstr = month
        if len(monthstr) == 1:
            monthstr = "0" + monthstr

        daystr: str
        if isinstance(day, int):
            daystr = str(day)
        else:
            daystr = day
        if len(daystr) == 1:
            daystr = "0" + daystr

        return cls.DATE_AND_TIMES_SIGIL + yearstr + "-" + monthstr + "-" + daystr + "T00:00:00/11"

    @classmethod
    def from_datetime(cls, d: dt.datetime, precision: typing.Optional[typing.Union[int, str]]=None)->str:
        if precision is None:
            return cls.DATE_AND_TIMES_SIGIL + dt.isoformat()
        else:
            return cls.DATE_AND_TIMES_SIGIL + dt.isoformat() + "/" + str(precision)
            

    @classmethod
    def lat_lon(cls, lat: typing.Union[int, float], lon: typing.Union[int, float])->str:
        return cls.LOCATION_COORDINATES_SIGIL + str(lat) + '/' + str(lon)



from kgtk.exceptions import KGTKException


def check_dependencies():
    pass_deps_check = True

    # requirements
    # TODO: The requirements.txt file won't be available in installed python package
    # print('checking requirements...')
    # import pkg_resources
    # with open('requirements.txt', 'r') as rf:
    #     requirements = pkg_resources.parse_requirements(rf.read())
    # for requirement in requirements:
    #     try:
    #         pkg_resources.require(str(requirement))
    #     except Exception as e:
    #         print(e)
    #         pass_deps_check = False

    # module level
    # print('checking other external dependencies...')
    if not check_deps_sqlite():
        pass_deps_check = False

    return pass_deps_check


### SQLite3

# The Python sqlite3 package and library is a builtin which does not
# require any special installation.  However some Kypher features such
# as explanation and fast, shell-based data import require a recent
# version of sqlite3 to be available as an executable shell command.
# The functions below check for those dependencies.

SQLITE_COMMAND = 'sqlite3'
SQLITE_MINIMUM_VERSION = '3.30'

def check_sqlite_command():
    """Return True if sqlite3 is available as a shell command and callable via 'sh'.
    """
    try:
        import sh
        sqlite3 = sh.Command(SQLITE_COMMAND)
        return True
    except:
        pass
    return False

def get_sqlite_version():
    try:
        import sh
        from packaging import version
        sqlite3 = sh.Command(SQLITE_COMMAND)
        return sqlite3('--version').stdout.split()[0].decode()
    except:
        return None

def check_sqlite_version():
    """Return True if sqlite3 is available as a shell command and has a recent enough version.
    """
    if check_sqlite_command():
        try:
            from packaging import version
            sqlite3_version = get_sqlite_version()
            return version.parse(sqlite3_version) >= version.parse(SQLITE_MINIMUM_VERSION)
        except:
            pass
    return False

def check_deps_sqlite():
    """Silently return True if sqlite3 is properly installed.
    Otherwise, print an explanation of the problem and return False.
    """
    if not check_sqlite_command():
        print('%s command is not installed or callable (please see manual)' % SQLITE_COMMAND)
        return False
    if not check_sqlite_version():
        print('%s version %s is too old, please upgrade to at least version %s (see manual)'
              % (SQLITE_COMMAND, get_sqlite_version(), SQLITE_MINIMUM_VERSION))
        return False
    return True



import json
from typing import Dict, List
from kgtk.knowledge_graph.schema import KGSchema
from kgtk.knowledge_graph.graph import Graph
from kgtk.knowledge_graph.subject import Subject
from kgtk.knowledge_graph.node import URI, Literal
from kgtk.utils import deprecated


class KnowledgeGraph(Graph):
    """
    This class is a knowledge graph object, provides API for user to construct their kg.
    Add field and value to the kg object, analysis on provenance
    """

    def __init__(self, schema: KGSchema, doc):
        super().__init__()
        self.origin_doc = doc
        self.schema = schema
        self._fork_namespace_manager()

    @deprecated()
    def add_value(self, field_name: str, value: object = None) -> None:
        """
        Add a value to knowledge graph.
        Input can either be a value or a json_path. If the input is json_path, the helper function _add_doc_value is
        called.
        If the input is a value, then it is handled

        Args:
            field_name: str, the field name in the knowledge graph
            value: the value to be added to the knowledge graph
        """
        if not self._ns.store.namespace(''):
            self.bind(None, 'http://isi.edu/default-ns/')
        obj = self.schema.field_type(field_name, value)
        if not obj:
            raise Exception()  # TODO: replace with a specific Exception
        self.add_triple(URI(self.origin_doc.doc_id), URI(field_name), obj)

    def _find_types(self, triples):
        """
        find type in root level
        :param triples:
        :return:
        """
        types = []
        for t in triples:
            s, p, o = t
            if self._is_rdf_type(p):
                if isinstance(o, Subject):
                    continue
                types.append(o)
        return types

    def add_subject(self, subjects, context=None):
        if not context:
            context = set([])
        s_types = self._find_types(subjects)

        for t in subjects:
            s, p, o = t
            o_types = []
            if isinstance(o, Subject) and o not in context:
                context.add(o)
                self.add_subject(o, context)
                o_types = self._find_types(o)

            if self.schema.is_valid(s_types, p, o_types):
                triple = self._convert_triple_rdflib((s, p, o))
                self._g.add(triple)

    @property
    def value(self) -> Dict:
        """
        Get knowledge graph object
        """
        g = {}
        for p, o in self._g.predicate_objects():
            _, property_ = self._ns.split_uri(p)
            if property_ not in g:
                g[property_] = list()
            g[property_].append({
                'key': self.create_key_from_value(o, property_),
                'value': o.toPython()
            })
        return g

    @deprecated()
    def get_values(self, field_name: str) -> List[object]:
        """
        Get a list of all the values of a field.
        """
        result = list()
        p = self.schema.parse_field(field_name)
        for o in self._g.objects(None, p):
            result.append(o.toPython())
        return result

    def create_key_from_value(self, value, field_name: str):
        key = self.schema.field_type(field_name, value)
        if isinstance(key, URI):
            return key
        if isinstance(key, str) or isinstance(key, Literal):
            key = str(key).strip().lower()
        return key

    def serialize(self, format='legacy', namespace_manager=None):
        if format == 'legacy':
            # Output DIG format
            return json.dumps(self.value)
        return super().serialize(format, namespace_manager)

    def _fork_namespace_manager(self):
        for prefix, ns in self.schema.ontology._ns.namespaces():
            self.bind(prefix, ns)

    def add_types(self, type_):
        s = Subject(URI(self.origin_doc.doc_id))
        p = URI('rdf:type')

        if not isinstance(type_, list):
            type_ = [type_]
        for a_type in type_:
            s.add_property(p, URI(a_type))
        self.add_subject(s)

    def validate(self):
        conforms, result_graph = self.schema.validate(self)
        return conforms, result_graph



"""
Generate mediawiki API json files from kgtk file

This command line tool will create three json files. Each will mimic the return of the following media wiki API for each entity existed in the kgtk file. 

"""

from argparse import Namespace, SUPPRESS
import typing
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def str2bool(v):
    import argparse
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def parser():
    """
    Initialize sub-parser.
    Parameters: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser
    """
    return {
        "help": "Generates mediawiki json responses from kgtk file",
        "description": "Generating json files that mimic mediawiki *wbgetentities* api call response. This tool assumes statements and qualifiers related to one entity will be bundled close as the `generate-wikidata-triples` function assumes. If this requirement is not met, please set `n` to a number LARGER than the total number of entities in the kgtk file",
    }

def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
        prop_file: str
    """
    from kgtk.utils.argparsehelpers import optional_bool

    # parser.add_argument(
    #    "-i",
    #    "--input-file",
    #    action="store",
    #    type=str,
    #    required = False,
    #    default="",
    #    help="set the path of the input kgtk file if not from standard input",
    #    dest="input_file",
    #)
    parser.add_input_file()

    # parser.add_argument(
    #    "-pf",
    #    "--property-file",
    #     action="store",
    #     type=str,
    #     required = False,
    #     default="NONE",
    #     help="path to the file which contains the property datatype mapping in kgtk format.",
    #     dest="prop_file",
    # )
    parser.add_input_file(who="the file which contains the property datatype mapping in kgtk format",
                          options=["-pf", "--property-file"],
                          dest="prop_file",
                          metavar="PROPERTY_FILE",
                          optional=True,
    )

    parser.add_argument(
        "-lp",
        "--label-property",
        action="store",
        type=str,
        default="label",
        required=False,
        help="property identifiers which will create labels, separated by comma','.",
        dest="labels",
    )
    parser.add_argument(
        "-ap",
        "--alias-property",
        action="store",
        type=str,
        required = False,
        default="alias",
        help="alias identifiers which will create labels, separated by comma','.",
        dest="aliases",
    )
    parser.add_argument(
        "-dp",
        "--description-property",
        action="store",
        type=str,
        required = False,
        default="description",
        help="description identifiers which will create labels, separated by comma','.",
        dest="descriptions",
    )
    parser.add_argument(
        "-pd",
        "--property-declaration-in-file",
        action="store",
        type=str2bool,
        required = False,
        default=False,
        help="whether to read properties from the input kgtk file. If set to yes, make sure the property declaration happens before its usage.  default=%(default)s",
        dest="prop_declaration",
    )
    parser.add_argument(
        "-pr",
        "--output-file-prefix",
        action="store",
        type=str,
        default = "kgtk",
        required = False,
        help="set the prefix of the output files. Default to `kgtk`",
        dest="output_prefix",
    )
    parser.add_argument(
        "-n",
        "--output-n-lines",
        action="store",
        type=int,
        required = False,
        default=1000,
        help="output json file when the corresponding dictionary size reaches n. Default to 1000",
        dest="n",
    )
    parser.add_argument(
        "-log",
        "--log-path",
        action="store",
        type=str,
        required = False,
        default="warning.log",
        help="set the path of the log file",
        dest="log_path",
    )
    parser.add_argument(
        "-w",
        "--warning",
        action="store",
        type=str2bool,
        required = False,
        default="no",
        help="if set to yes, warn various kinds of exceptions and mistakes and log them to a log file with line number in input file, rather than stopping. logging",
        dest="warning",
    )
    parser.add_argument(
        "-r",
        "--rank",
        action="store",
        type=bool,
        required = False,
        default=False,
        help="Whether the input file contains a rank column. Please refer to the `import_wikidata` command for the header information. Default to False, then all the ranks will be `normal`, therefore `NormalRank`.",
        dest="has_rank",
    )
    parser.add_argument(
        "--error-action",
        action="store",
        type=str,
        required = False,
        default="log",
        choices=['ignore', 'log', 'raise'],
        help="When errors occur, either ignore them ('ignore'), log them (`log`), or raise an exception (`raise`). Default='%(default)s'.",
        dest="error_action",
    )
    parser.add_argument(
        "-pl",
        "--property-declaration-label",
        action="store",
        type=str,
        required=False,
        default="data_type",
        help="The edge label in a property file that indicates a property declaration. default='%(default)s'",
        dest="property_declaration_label",
    )

    parser.add_argument(
        "-fp",
        "--filter-prop-file",
        dest="filter_prop_file",
        help="If true and a property file has been specified, filter the prop file, processing only edges with the property declaration label. (default=%(default)s)",
        type=optional_bool, nargs='?', const=True, default=True, metavar="True/False",
    )

    parser.add_argument(
        "-ip",
        "--ignore-property-declarations-in-file",
        dest="ignore_property_declarations_in_file",
        help="If true, ignore input edges with the property declaration label. (default=%(default)s)",
        type=optional_bool, nargs='?', const=True, default=True, metavar="True/False",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        help="If true, provide additional feedback. (default=%(default)s)",
        type=optional_bool, nargs='?', const=True, default=False, metavar="True/False",
    )

def run(
    input_file: KGTKFiles,
    prop_file: KGTKFiles,

    labels: str,
    aliases: str,
    descriptions: str,
    prop_declaration:bool,
    output_prefix: str,
    n: int,
    log_path: str,
    warning: bool,
    has_rank:bool,
    error_action: str,
    property_declaration_label: str,
    ignore_property_declarations_in_file: bool,
    filter_prop_file: bool,
    verbose: bool,
):
    # import modules locally
    from pathlib import Path
    from kgtk.generator import JsonGenerator
    import sys
    import gzip
    from kgtk.exceptions import KGTKException
    
    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    prop_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_input_file(prop_file, who="KGTK prop file")

    generator = JsonGenerator(
        input_file = input_kgtk_file,
        prop_file=prop_kgtk_file,
        label_set=labels,
        alias_set=aliases,
        description_set=descriptions,
        output_prefix = output_prefix,
        n = n,
        log_path = log_path,
        warning = warning,
        prop_declaration = prop_declaration,
        has_rank = has_rank,
        error_action = error_action,
        property_declaration_label=property_declaration_label,
        ignore_property_declarations_in_file = ignore_property_declarations_in_file,
        filter_prop_file = filter_prop_file,
        verbose = verbose,
    )
    generator.process()



"""
Cat multuple KGTK file together.

TODO: Need output file mode.

"""

from argparse import ArgumentParser
import attr
import os
from pathlib import Path
import sh  # type: ignore
import sys
import typing

from kgtk.cli_entry import progress_startup
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.join.kgtkmergecolumns import KgtkMergeColumns
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkCat():
    DEFAULT_PURE_PYTHON: bool = False
    DEFAULT_FAST_COPY_MIN_SIZE: int = 10000
    DEFAULT_BASH_COMMAND: str = "bash"
    DEFAULT_BZIP2_COMMAND: str = "bzip2"
    DEFAULT_CAT_COMMAND: str = "cat"
    DEFAULT_GZIP_COMMAND: str = "gzip"
    DEFAULT_TAIL_COMMAND: str = "tail"
    DEFAULT_XZ_COMMAND: str = "xz"

    input_file_paths: typing.List[Path] = attr.ib()
    output_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    # When supplied, rename all output columns.
    output_column_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    # When supplied, renames selected colums: old names.
    old_column_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    # When supplied, renames selected colums: new names.
    new_column_names: typing.Optional[typing.List[str]] = \
        attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                 iterable_validator=attr.validators.instance_of(list))),
                default=None)

    no_output_header: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    pure_python: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_PURE_PYTHON)
    fast_copy_min_size: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_FAST_COPY_MIN_SIZE)

    bash_command: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_BASH_COMMAND)
    bzip2_command: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_BZIP2_COMMAND)
    cat_command: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_CAT_COMMAND)
    gzip_command: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_GZIP_COMMAND)
    tail_command: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_TAIL_COMMAND)
    xz_command: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_XZ_COMMAND)

    # TODO: find working validators:
    reader_options: typing.Optional[KgtkReaderOptions] = attr.ib(default=None)
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    output_format: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None) # TODO: use an enum

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def process(self):
        kmc: KgtkMergeColumns = KgtkMergeColumns()

        # Is the output file an edge file, a node file, or unknown?
        is_edge_file: bool = False
        is_node_file: bool = False

        krs: typing.List[KgtkReader] = [ ]
        kr: KgtkReader
        idx: int

        if self.verbose:
            print("Starting kgtkcat pid=%d" % (os.getpid()), file=self.error_file, flush=True)

        if self.verbose:
            print("Opening the %d input files." % len(self.input_file_paths), file=self.error_file, flush=True)

        use_system_copy: bool = not self.pure_python
        if self.output_format is not None and self.output_format != KgtkWriter.OUTPUT_FORMAT_KGTK:
            # TODO: OK if the input and output formats are both CSV with headers.
            use_system_copy = False
        initial_column_names: typing.Optional[typing.List[str]] = None

        saw_stdin: bool = False
        input_file_path: Path
        for idx, input_file_path in enumerate(self.input_file_paths):
            if str(input_file_path) == "-":
                if saw_stdin:
                    raise ValueError("Duplicate standard input file %d" % (idx + 1))
                else:
                    saw_stdin = False
                if self.verbose:
                    print("Opening file %d: standard input" % (idx + 1), file=self.error_file, flush=True)
            else:
                if self.verbose:
                    print("Opening file %d: %s" % (idx + 1, str(input_file_path)), file=self.error_file, flush=True)

            kr = KgtkReader.open(input_file_path,
                                 who="input " + str(idx + 1),
                                 options=self.reader_options,
                                 value_options = self.value_options,
                                 error_file=self.error_file,
                                 verbose=self.verbose,
                                 very_verbose=self.very_verbose,
            )
            krs.append(kr)

            # Unless directed otherwise, do not merge edge files with node
            # files.  If options.mode == KgtkReaderMode.NONE, then neither
            # kr.is_edge_file nor kr.is_node_file will be set and the
            # consistency check will be skipped.
            if kr.is_edge_file:
                if is_node_file:
                    # Close the open files before raising the exception.
                    #
                    # TODO: Use a try..finally block to ensure these files are closed.
                    for kr2 in krs:
                        kr2.close()
                    raise ValueError("Cannot merge an edge file to a node file: %s" % input_file_path)
                if is_edge_file == False and self.verbose:
                    print("The output file will be an edge file.", file=self.error_file, flush=True)
                is_edge_file = True
            elif kr.is_node_file:
                if is_edge_file:
                    # Close the open files before raising the exception.
                    #
                    # TODO: Use a try..finally block to ensure these files are closed.
                    for kr2 in krs:
                        kr2.close()
                    raise ValueError("Cannot merge a node file to an edge file: %s" % input_file_path)
                if is_node_file == False and self.verbose:
                    print("The output file will be an node file.", file=self.error_file, flush=True)
                is_node_file = True

            if self.verbose or self.very_verbose:
                print("Mapping the %d column names in %s." % (len(kr.column_names), input_file_path), file=self.error_file, flush=True)
            if self.very_verbose:
                print(" ".join(kr.column_names), file=self.error_file, flush=True)
            new_column_names: typing.List[str] =  kmc.merge(kr.column_names)
            if self.very_verbose:
                print(" ".join(new_column_names), file=self.error_file, flush=True)

            # Can we still use the system copy?
            if not kr.use_fast_path:
                use_system_copy = False
            if kr.options.force_column_names is not None:
                use_system_copy = False
            if kr.options.supply_missing_column_names:
                use_system_copy = False
            if kr.options.no_input_header:
                use_system_copy = False
            if kr.options.number_of_columns is not None:
                use_system_copy = False
            if kr.options.require_column_names is not None:
                # This constraint could be removed.
                use_system_copy = False
            if kr.options.no_additional_columns:
                # This constraint could be removed.
                use_system_copy = False
            if not kr.rewindable:
                use_system_copy = False
            if initial_column_names is None:
                initial_column_names = kr.column_names.copy()
            else:
                # TODO: Account for coumn name aliases.
                if initial_column_names != kr.column_names:
                    use_system_copy = False

        if self.verbose or self.very_verbose:
            print("There are %d merged columns." % len(kmc.column_names), file=self.error_file, flush=True)
        if self.very_verbose:
            print(" ".join(kmc.column_names), file=self.error_file, flush=True)
            
        if self.output_column_names is not None:
            if self.verbose:
                print("There are %d new output column names." % len(self.output_column_names), file=self.error_file, flush=True)
            if len(self.output_column_names) != len(kmc.column_names):
                # Close the open files before raising the exception.
                #
                # TODO: Use a try..finally block to ensure these files are closed.
                for kr2 in krs:
                    kr2.close()
                raise ValueError("There are %d merged columns, but %d output column names." % (len(kmc.column_names), len(self.output_column_names)))

        if use_system_copy:
            # TODO: restructure this code for better readability.
            if self.verbose:
                print("Using the system commands for fast copies.", file=self.error_file, flush=True)
            copied_column_names: typing.List[str] = initial_column_names
            if self.output_column_names is not None:
                copied_column_names = self.output_column_names
            if self.do_system_copy(krs, copied_column_names):
                return
        progress_startup()

        output_mode: KgtkWriter.Mode = KgtkWriter.Mode.NONE
        if is_edge_file:
            output_mode = KgtkWriter.Mode.EDGE
            if self.verbose:
                print("Opening the output edge file: %s" % str(self.output_path), file=self.error_file, flush=True)
        elif is_node_file:
            output_mode = KgtkWriter.Mode.NODE
            if self.verbose:
                print("Opening the output node file: %s" % str(self.output_path), file=self.error_file, flush=True)
        else:
            if self.verbose:
                print("Opening the output file: %s" % str(self.output_path), file=self.error_file, flush=True)


        ew: KgtkWriter = KgtkWriter.open(kmc.column_names,
                                         self.output_path,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         use_mgzip=self.reader_options.use_mgzip, # Hack!
                                         mgzip_threads=self.reader_options.mgzip_threads, # Hack!
                                         gzip_in_parallel=False,
                                         mode=output_mode,
                                         output_format=self.output_format,
                                         output_column_names=self.output_column_names,
                                         old_column_names=self.old_column_names,
                                         new_column_names=self.new_column_names,
                                         no_header=self.no_output_header,
                                         error_file=self.error_file,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        output_data_lines: int = 0
        for idx, kr in enumerate(krs):
            if kr.file_path is None:
                # This shouldn't happen because we constrined all
                # input_file_path elements to be not None.  However,
                # checking here keeps mypy happy.
                #
                # TODO: throw a better exception.
                #
                # Close the open files before raising the exception.
                #
                # TODO: Use a try..finally block to ensure these files are closed.
                for kr2 in krs:
                    kr2.close()
                raise ValueError("Missing file path.")
            input_file_path = kr.file_path
            if self.verbose:
                print("Copying data from file %d: %s" % (idx + 1, input_file_path), file=self.error_file, flush=True)

            input_data_lines: int = ew.copyfile(kr, new_column_names=kmc.new_column_name_lists[idx])
            output_data_lines += input_data_lines

            if self.verbose:
                print("Read %d data lines from file %d: %s" % (input_data_lines, idx + 1, input_file_path), file=self.error_file, flush=True)
        
        if self.verbose:
            print("Wrote %d lines total from %d files" % (output_data_lines, len(krs)), file=self.error_file, flush=True)

        # Close the open files.
        ew.close()
        for kr2 in krs:
            kr2.close()

    def safe_filename(self, file_path: Path) -> str:
        """Convert a path into a quoted and escaped filename that is safe for bash."""
        filename: str = str(file_path)
        if filename.startswith('-') and filename != '-':
            # Prevent filenames that begin with dash from being treated as
            # command options.  Strictly speaking, this is not a bash problem
            # per se.
            filename = './' + filename
        return "'" + filename.replace("'", "'\\''") + "'"

    def do_system_copy(self,
                       krs: typing.List[KgtkReader],
                       column_names: typing.List[str]) -> bool:

        # TODO: Support numbered FDs as input files.
        input_file_path: str
        for input_file_path in self.input_file_paths:
            filename: str = str(input_file_path)
            if filename.startswith("<"):
                if verbose:
                    print(("Cannot use numbered FDs with system tools yet (%s)."
                           % repr(filename)),
                          file=self.error_file, flush=True)
                return False

        # Sum the the sizes of the input files.  Skip the fast copy if
        # the total size is too small.
        total_input_file_size: int = 0
        for input_file_path in self.input_file_paths:
            total_input_file_size += input_file_path.stat().st_size
        if total_input_file_size < self.fast_copy_min_size:
            if self.verbose:
                print(("The total file size (%d) is less than the minimum "
                       + "for fast copies (%d).") % (total_input_file_size,
                                                     self.fast_copy_min_size),
                      file=self.error_file, flush=True)
            return False  # Take the slow path.
        if self.verbose:
            print(("The total file size (%d) meets the minimum "
                   + "for fast copies (%d).") % (total_input_file_size,
                                                 self.fast_copy_min_size),
                  file=self.error_file, flush=True)

        # Close the open files.
        for kr2 in krs:
            kr2.close()

        cmd: str = "("

        idx: int
        for idx, input_file_path in enumerate(self.input_file_paths):
            input_suffix: str = input_file_path.suffix.lower()
            if idx == 0:
                if input_suffix in [".gz", ".z"]:
                    cmd += " " + self.gzip_command + " --decompress --stdout "
                elif input_suffix in [".bz2", ".bz"]:
                    cmd += " " + self.bzip2_command + " --decompress --stdout "
                elif input_suffix in [".xz", ".lzma"]:
                    cmd += " " + self.xz_command + " --decompress --stdout "
                else:
                    cmd += " " + self.cat_command + " "

                cmd += self.safe_filename(input_file_path)

            else:
                cmd += " && "
                if input_suffix in [".gz", ".z"]:
                    cmd += (self.gzip_command + " --decompress --stdout "
                            + self.safe_filename(input_file_path)
                            + " | " + self.tail_command + " -n +2")
                elif input_suffix in [".bz2", ".bz"]:
                    cmd += (self.bzip2_command + " --decompress --stdout "
                            + self.safe_filename(input_file_path)
                            + " | " + self.tail_command + " -n +2")
                elif input_suffix in [".xz", ".lzma"]:
                    cmd += (self.xz_command + " --decompress --stdout "
                            + self.safe_filename(input_file_path)
                            + " | " + self.tail_command + " -n +2")
                else:
                    cmd += self.tail_command + " -n +2 " + self.safe_filename(input_file_path)

        cmd += " )"
        if self.output_path is not None and str(self.output_path) != "-":
            output_suffix: str = self.output_path.suffix.lower()
            if input_suffix in [".gz", ".z"]:
                cmd += " | " + self.gzip_command
            elif input_suffix in [".bz2", ".bz"]:
                cmd += " | " + self.bzip2_command
            elif input_suffix in [".xz", ".lzma"]:
                cmd += " | " + self.xz_command

            cmd += " "
            if str(self.output_path).startswith(">"):
                cmd += str(self.output_path)
            else:
                cmd += '>' + self.safe_filename(self.output_path)

        if self.verbose:
            print("system command: %s" % repr(cmd), file=self.error_file, flush=True)

        sh_bash = sh.Command(self.bash_command)
        cmd_proc = sh_bash("-c", cmd, _out=sys.stdout, _err=sys.stderr,
                           bg=True, _bg_exc=False, _internal_bufsize=1)

        if self.verbose:
            print("\nRunning the cat script (pid=%d)." % cmd_proc.pid, file=self.error_file, flush=True)
        progress_startup(pid=cmd_proc.pid)

        if self.verbose:
            print("\nWaiting for the cat command to complete.\n", file=self.error_file, flush=True)
        cmd_proc.wait()

        return True

        
def main():
    """
    Test the KGTK file concatenator.
    """
    parser = ArgumentParser()
    parser.add_argument(dest="input_file_paths", help="The KGTK files to concatenate", type=Path, nargs='+')
    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s)", type=Path, default="-")

    parser.add_argument(      "--output-format", dest="output_format", help="The file format (default=kgtk)", type=str,
                              choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    parser.add_argument(      "--output-columns", dest="output_column_names", help="Rename all output columns. (default=%(default)s)", type=str, nargs='+')
    parser.add_argument(      "--old-columns", dest="old_column_names", help="Rename seleted output columns: old names. (default=%(default)s)", type=str, nargs='+')
    parser.add_argument(      "--new-columns", dest="new_column_names", help="Rename seleted output columns: new names. (default=%(default)s)", type=str, nargs='+')

    KgtkReader.add_debug_arguments(parser, expert=True)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=True)
    KgtkValueOptions.add_arguments(parser, expert=True)

    args = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args)
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        input_files: typing.List[str] = [ ]
        input_file: Path
        for input_file in args.input_file_paths:
            input_files.append(str(input_file))
        print("input: %s" % " ".join(input_files), file=error_file, flush=True)
        print("--output-file=%s" % args.output_file_path, file=error_file, flush=True)
        if args.output_format is not None:
            print("--output-format=%s" % args.output_format, file=error_file, flush=True)
        if args.output_column_names is not None:
            print("--output-columns=%s" %" ".join(args.output_column_names), file=error_file, flush=True)
        if args.old_column_names is not None:
            print("--old-columns=%s" %" ".join(args.old_column_names), file=error_file, flush=True)
        if args.new_column_names is not None:
            print("--new-columns=%s" %" ".join(args.new_column_names), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)

    kc: KgtkCat = KgtkCat(input_file_paths=args.input_file_paths,
                          output_path=args.output_file_path,
                          output_format=args.output_format,
                          output_column_names=args.output_column_names,
                          old_column_names=args.old_column_names,
                          new_column_names=args.new_column_names,
                          reader_options=reader_options,
                          value_options=value_options,
                          error_file=error_file,
                          verbose=args.verbose,
                          very_verbose=args.very_verbose)

    kc.process()

if __name__ == "__main__":
    main()




import unittest
import sys
from kgtk.cli_entry import cli_entry
import tempfile
import pandas as pd
import shutil


class TestEmbedding(unittest.TestCase):
    def setUp(self) -> None:
        self.edge_file_path = 'data/text_embedding.edges.tsv'
        self.true_output_file_path = 'data/text_embedding.output.tsv'
        self.true_output_w2v_file_path = 'data/text_embedding.output.w2v.txt'
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_vector(self):
        cli_entry("kgtk", "text-embedding", "-i", self.edge_file_path,
                  "--model", "roberta-large-nli-mean-tokens",
                  "--debug",
                  "--sentence-property", "sentence",
                  "--output-property", "text_embedding",
                  "--batch-size", "-1",
                  "--output-data-format", "kgtk",
                  "-o", f'{self.temp_dir}/text-embedding.tsv')

        df_r = pd.read_csv(self.true_output_file_path, sep='\t')

        df = pd.read_csv(f'{self.temp_dir}/text-embedding.tsv', sep='\t')
        self.assertEqual(len(df), len(df_r))
        self.assertEqual(list(df.columns), list(df_r.columns))

        for temprow1, temprow2 in zip(df_r.iterrows(), df.iterrows()):
            row1, row2 = temprow1[1], temprow2[1]
            self.assertEqual(row1["node1"], row2["node1"])
            self.assertEqual(row1["label"], row2["label"])
            self.assertTrue(len(row2["node2"]) != 0)
            self.assertEqual(len(row1["node2"].split(",")), len(row2["node2"].split(",")))

    def test_w2v_vector(self):
        cli_entry("kgtk", "text-embedding", "-i", self.edge_file_path,
                  "--model", "roberta-large-nli-mean-tokens",
                  "--debug",
                  "--sentence-property", "sentence",
                  "--output-property", "text_embedding",
                  "--batch-size", "-1",
                  "--output-data-format", "w2v",
                  "-o", f'{self.temp_dir}/text-embedding.txt')

        with open(f'{self.temp_dir}/text-embedding.txt') as f1:
            o_lines = f1.readlines()

        with open(self.true_output_w2v_file_path) as f2:
            truth_lines = f2.readlines()
        self.assertEqual(len(o_lines), len(truth_lines))
        self.assertEqual(o_lines[0], truth_lines[0])

        for line1, line2 in zip(o_lines[1:], truth_lines[1:]):
            vals1 = line1.strip().split(" ")
            vals2 = line2.strip().split(" ")
            self.assertEqual(vals1[0], vals2[0])
            self.assertEqual(len(vals1), len(vals2))
